# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Basic resource implementations

A resource in URL / CoAP / REST terminology is the thing identified by a URI.

Here, a :class:`.Resource` is the place where server functionality is
implemented. In many cases, there exists one persistent Resource object for a
given resource (eg. a ``TimeResource()`` is responsible for serving the
``/time`` location). On the other hand, an aiocoap server context accepts only
one thing as its serversite, and that is a Resource too (typically of the
:class:`Site` class).

Resources are most easily implemented by deriving from :class:`.Resource` and
implementing ``render_get``, ``render_post`` and similar coroutine methods.
Those take a single request message object and must return a
:class:`aiocoap.Message` object or raise an
:class:`.error.RenderableError` (eg. ``raise UnsupportedMediaType()``).

To serve more than one resource on a site, use the :class:`Site` class to
dispatch requests based on the Uri-Path header.
"""

import hashlib
import warnings

from . import message
from . import meta
from . import error
from . import interfaces
from . import numbers
from .pipe import Pipe

def hashing_etag(request, response):
    """Helper function for render_get handlers that allows them to use ETags based
    on the payload's hash value

    Run this on your request and response before returning from render_get; it is
    safe to use this function with all kinds of responses, it will only act on
    2.05 Content messages (and those with no code set, which defaults to that
    for GET requests). The hash used are the first 8 bytes of the sha1 sum of
    the payload.

    Note that this method is not ideal from a server performance point of view
    (a file server, for example, might want to hash only the stat() result of a
    file instead of reading it in full), but it saves bandwith for the simple
    cases.

    >>> from aiocoap import *
    >>> req = Message(code=GET)
    >>> hash_of_hello = b'\\xaa\\xf4\\xc6\\x1d\\xdc\\xc5\\xe8\\xa2'
    >>> req.opt.etags = [hash_of_hello]
    >>> resp = Message(code=CONTENT)
    >>> resp.payload = b'hello'
    >>> hashing_etag(req, resp)
    >>> resp                                            # doctest: +ELLIPSIS
    <aiocoap.Message at ... 2.03 Valid ... 1 option(s)>
    """

    if response.code != numbers.codes.CONTENT and response.code is not None:
        return

    response.opt.etag = hashlib.sha1(response.payload).digest()[:8]
    if request.opt.etags is not None and response.opt.etag in request.opt.etags:
        response.code = numbers.codes.VALID
        response.payload = b''

class _ExposesWellknownAttributes:
    def get_link_description(self):
        # FIXME which formats are acceptable, and how much escaping and
        # list-to-separated-string conversion needs to happen here
        ret = {}
        if hasattr(self, 'ct'):
            ret['ct'] = str(self.ct)
        if hasattr(self, 'rt'):
            ret['rt'] = self.rt
        if hasattr(self, 'if_'):
            ret['if'] = self.if_
        return ret

class Resource(_ExposesWellknownAttributes, interfaces.Resource):
    """Simple base implementation of the :class:`interfaces.Resource`
    interface

    The render method delegates content creation to ``render_$method`` methods
    (``render_get``, ``render_put`` etc), and responds appropriately to
    unsupported methods. Those messages may return messages without a response
    code, the default render method will set an appropriate successful code
    ("Content" for GET/FETCH, "Deleted" for DELETE, "Changed" for anything
    else). The render method will also fill in the request's no_response code
    into the response (see :meth:`.interfaces.Resource.render`) if none was
    set.

    Moreover, this class provides a ``get_link_description`` method as used by
    .well-known/core to expose a resource's ``.ct``, ``.rt`` and ``.if_``
    (alternative name for ``if`` as that's a Python keyword) attributes.
    Details can be added by overriding the method to return a more
    comprehensive dictionary, and resources can be hidden completely by
    returning None.
    """

    async def needs_blockwise_assembly(self, request):
        return True

    async def render(self, request):
        if not request.code.is_request():
            raise error.UnsupportedMethod()
        m = getattr(self, 'render_%s' % str(request.code).lower(), None)
        if not m:
            raise error.UnallowedMethod()

        response = await m(request)

        if response is message.NoResponse:
            warnings.warn("Returning NoResponse is deprecated, please return a"
                          " regular response with a no_response option set.",
                          DeprecationWarning)
            response = message.Message(no_response=26)

        if response.code is None:
            if request.code in (numbers.codes.GET, numbers.codes.FETCH):
                response_default = numbers.codes.CONTENT
            elif request.code == numbers.codes.DELETE:
                response_default = numbers.codes.DELETED
            else:
                response_default = numbers.codes.CHANGED
            response.code = response_default

        if response.opt.no_response is None:
            response.opt.no_response = request.opt.no_response

        return response

    async def render_to_pipe(self, request: Pipe):
        # Silence the deprecation warning
        if isinstance(self, interfaces.ObservableResource):
            # See interfaces.Resource.render_to_pipe
            return await interfaces.ObservableResource._render_to_pipe(self, request)
        return await interfaces.Resource._render_to_pipe(self, request)

class ObservableResource(Resource, interfaces.ObservableResource):
    def __init__(self):
        super(ObservableResource, self).__init__()
        self._observations = set()

    async def add_observation(self, request, serverobservation):
        self._observations.add(serverobservation)
        def _cancel(self=self, obs=serverobservation):
            self._observations.remove(serverobservation)
            self.update_observation_count(len(self._observations))
        serverobservation.accept(_cancel)
        self.update_observation_count(len(self._observations))

    def update_observation_count(self, newcount):
        """Hook into this method to be notified when the number of observations
        on the resource changes."""

    def updated_state(self, response=None):
        """Call this whenever the resource was updated, and a notification
        should be sent to observers."""

        for o in self._observations:
            o.trigger(response)

    def get_link_description(self):
        link = super(ObservableResource, self).get_link_description()
        link['obs'] = None
        return link

    async def render_to_pipe(self, request: Pipe):
        # Silence the deprecation warning
        return await interfaces.ObservableResource._render_to_pipe(self, request)

def link_format_to_message(request, linkformat,
        default_ct=numbers.ContentFormat.LINKFORMAT):
    """Given a LinkFormat object, render it to a response message, picking a
    suitable conent format from a given request.

    It returns a Not Acceptable response if something unsupported was queried.

    It makes no attempt to modify the URI reference literals encoded in the
    LinkFormat object; they have to be suitably prepared by the caller."""

    ct = request.opt.accept
    if ct is None:
        ct = default_ct

    if ct == numbers.ContentFormat.LINKFORMAT:
        payload = str(linkformat).encode('utf8')
    else:
        return message.Message(code=numbers.NOT_ACCEPTABLE)

    return message.Message(payload=payload, content_format=ct)

# Convenience attribute to set as ct on resources that use
# link_format_to_message as their final step in the request handler
link_format_to_message.supported_ct = " ".join(str(int(x)) for x in (
        numbers.ContentFormat.LINKFORMAT,
        ))

class WKCResource(Resource):
    """Read-only dynamic resource list, suitable as .well-known/core.

    This resource renders a link_header.LinkHeader object (which describes a
    collection of resources) as application/link-format (RFC 6690).

    The list to be rendered is obtained from a function passed into the
    constructor; typically, that function would be a bound
    Site.get_resources_as_linkheader() method.

    This resource also provides server `implementation information link`_;
    server authors are invited to override this by passing an own URI as the
    `impl_info` parameter, and can disable it by passing None.

    .. _`implementation information link`: https://tools.ietf.org/html/draft-bormann-t2trg-rel-impl-00"""

    ct = link_format_to_message.supported_ct

    def __init__(self, listgenerator, impl_info=meta.library_uri, **kwargs):
        super().__init__(**kwargs)
        self.listgenerator = listgenerator
        self.impl_info = impl_info

    async def render_get(self, request):
        links = self.listgenerator()

        if self.impl_info is not None:
            from .util.linkformat import Link
            links.links = links.links + [Link(href=self.impl_info, rel="impl-info")]

        filters = []
        for q in request.opt.uri_query:
            try:
                k, v = q.split('=', 1)
            except ValueError:
                continue # no =, not a relevant filter

            if v.endswith('*'):
                def matchexp(x, v=v):
                    return x.startswith(v[:-1])
            else:
                def matchexp(x, v=v):
                    return x == v

            if k in ('rt', 'if', 'ct'):
                filters.append(lambda link: any(matchexp(part) for part in (" ".join(getattr(link, k, ()))).split(" ")))
            elif k in ('href',): # x.href is single valued
                filters.append(lambda link: matchexp(getattr(link, k)))
            else:
                filters.append(lambda link: any(matchexp(part) for part in getattr(link, k, ())))

        while filters:
            links.links = filter(filters.pop(), links.links)
        links.links = list(links.links)

        response = link_format_to_message(request, links)

        if request.opt.uri_query and not links.links and \
                request.remote.is_multicast_locally:
            if request.opt.no_response is None:
                # If the filter does not match, multicast requests should not
                # be responded to -- that's equivalent to a "no_response on
                # 2.xx" option.
                response.opt.no_response = 0x02

        return response

class PathCapable:
    """Class that indicates that a resource promises to parse the uri_path
    option, and can thus be given requests for :meth:`.render`-ing that
    contain a uri_path"""

class Site(interfaces.ObservableResource, PathCapable):
    """Typical root element that gets passed to a :class:`Context` and contains
    all the resources that can be found when the endpoint gets accessed as a
    server.

    This provides easy registration of statical resources. Add resources at
    absolute locations using the :meth:`.add_resource` method.

    For example, the site at

    >>> site = Site()
    >>> site.add_resource(["hello"], Resource())

    will have requests to </hello> rendered by the new resource.

    You can add another Site (or another instance of :class:`PathCapable`) as
    well, those will be nested and integrally reported in a WKCResource. The
    path of a site should not end with an empty string (ie. a slash in the URI)
    -- the child site's own root resource will then have the trailing slash
    address.  Subsites can not have link-header attributes on their own (eg.
    `rt`) and will never respond to a request that does not at least contain a
    single slash after the the given path part.

    For example,

    >>> batch = Site()
    >>> batch.add_resource(["light1"], Resource())
    >>> batch.add_resource(["light2"], Resource())
    >>> batch.add_resource([], Resource())
    >>> s = Site()
    >>> s.add_resource(["batch"], batch)

    will have the three created resources rendered at </batch/light1>,
    </batch/light2> and </batch/>.

    If it is necessary to respond to requests to </batch> or report its
    attributes in .well-known/core in addition to the above, a non-PathCapable
    resource can be added with the same path. This is usually considered an odd
    design, not fully supported, and for example doesn't support removal of
    resources from the site.
    """

    def __init__(self):
        self._resources = {}
        self._subsites = {}

    async def needs_blockwise_assembly(self, request):
        try:
            child, subrequest = self._find_child_and_pathstripped_message(request)
        except KeyError:
            return True
        else:
            return await child.needs_blockwise_assembly(subrequest)

    def _find_child_and_pathstripped_message(self, request):
        """Given a request, find the child that will handle it, and strip all
        path components from the request that are covered by the child's
        position within the site. Returns the child and a request with a path
        shortened by the components in the child's path, or raises a
        KeyError.

        While producing stripped messages, this adds a ._original_request_uri
        attribute to the messages which holds the request URI before the
        stripping is started. That allows internal components to access the
        original URI until there is a variation of the request API that allows
        accessing this in a better usable way."""

        original_request_uri = getattr(request, '_original_request_uri',
                request.get_request_uri(local_is_server=True))

        if request.opt.uri_path in self._resources:
            stripped = request.copy(uri_path=())
            stripped._original_request_uri = original_request_uri
            return self._resources[request.opt.uri_path], stripped

        if not request.opt.uri_path:
            raise KeyError()

        remainder = [request.opt.uri_path[-1]]
        path = request.opt.uri_path[:-1]
        while path:
            if path in self._subsites:
                res = self._subsites[path]
                if remainder == [""]:
                    # sub-sites should see their root resource like sites
                    remainder = []
                stripped = request.copy(uri_path=remainder)
                stripped._original_request_uri = original_request_uri
                return res, stripped
            remainder.insert(0, path[-1])
            path = path[:-1]
        raise KeyError()

    async def render(self, request):
        try:
            child, subrequest = self._find_child_and_pathstripped_message(request)
        except KeyError:
            raise error.NotFound()
        else:
            return await child.render(subrequest)

    async def add_observation(self, request, serverobservation):
        try:
            child, subrequest = self._find_child_and_pathstripped_message(request)
        except KeyError:
            return

        try:
            await child.add_observation(subrequest, serverobservation)
        except AttributeError:
            pass

    def add_resource(self, path, resource):
        if isinstance(path, str):
            raise ValueError("Paths should be tuples or lists of strings")
        if isinstance(resource, PathCapable):
            self._subsites[tuple(path)] = resource
        else:
            self._resources[tuple(path)] = resource

    def remove_resource(self, path):
        try:
            del self._subsites[tuple(path)]
        except KeyError:
            del self._resources[tuple(path)]

    def get_resources_as_linkheader(self):
        from .util.linkformat import Link, LinkFormat

        links = []

        for path, resource in self._resources.items():
            if hasattr(resource, "get_link_description"):
                details = resource.get_link_description()
            else:
                details = {}
            if details is None:
                continue
            lh = Link('/' + '/'.join(path), **details)

            links.append(lh)

        for path, resource in self._subsites.items():
            if hasattr(resource, "get_resources_as_linkheader"):
                for l in resource.get_resources_as_linkheader().links:
                    links.append(Link('/' + '/'.join(path) + l.href, l.attr_pairs))
        return LinkFormat(links)

    async def render_to_pipe(self, request: Pipe):
        try:
            child, subrequest = self._find_child_and_pathstripped_message(request.request)
        except KeyError:
            raise error.NotFound()
        else:
            # FIXME consider carefully whether this switching-around is good.
            # It probably is.
            request.request = subrequest
            return await child.render_to_pipe(request)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module assists in creating OSCORE servers by proving a wrapper around a
:class:aiocoap.resource.Site. It enforces no access control, but just indicates
to the resources whether a client is authenticated by setting the request's
remote property adaequately.
"""

import logging

import aiocoap
from aiocoap import interfaces
from aiocoap import oscore, error
import aiocoap.pipe
from .numbers.codes import FETCH, POST

from aiocoap.transports.oscore import OSCOREAddress

class OscoreSiteWrapper(interfaces.Resource):
    def __init__(self, inner_site, server_credentials):
        self.log = logging.getLogger('oscore-site')

        self._inner_site = inner_site
        self.server_credentials = server_credentials

    async def render(self, request):
        raise RuntimeError("OscoreSiteWrapper can only be used through the render_to_pipe interface")

    async def needs_blockwise_assembly(self, request):
        raise RuntimeError("OscoreSiteWrapper can only be used through the render_to_pipe interface")

    # FIXME: should there be a get_resources_as_linkheader that just forwards
    # all the others and indicates ;osc everywhere?

    async def render_to_pipe(self, pipe):
        request = pipe.request

        try:
            unprotected = oscore.verify_start(request)
        except oscore.NotAProtectedMessage:
            # ie. if no object_seccurity present
            await self._inner_site.render_to_pipe(pipe)
            return

        if request.code not in (FETCH, POST):
            raise error.MethodNotAllowed

        try:
            sc = self.server_credentials.find_oscore(unprotected)
        except KeyError:
            if request.mtype == aiocoap.CON:
                raise error.Unauthorized("Security context not found")
            else:
                return

        try:
            unprotected, seqno = sc.unprotect(request)
        except error.RenderableError as e:
            # Note that this is flying out of the unprotection (ie. the
            # security context), which is trusted to not leak unintended
            # information in unencrypted responses. (By comparison, a
            # renderable exception flying out of a user
            # render_to_pipe could only be be rendered to a
            # protected message, and we'd need to be weary of rendering errors
            # during to_message as well).
            #
            # Note that this clause is not a no-op: it protects the 4.01 Echo
            # recovery exception (which is also a ReplayError) from being
            # treated as such.
            raise e
        # The other errors could be ported thee but would need some better NoResponse handling.
        except oscore.ReplayError:
            if request.mtype == aiocoap.CON:
                pipe.add_response(
                        aiocoap.Message(code=aiocoap.UNAUTHORIZED, max_age=0, payload=b"Replay detected"),
                        is_last=True)
            return
        except oscore.DecodeError:
            if request.mtype == aiocoap.CON:
                raise error.BadOption("Failed to decode COSE")
            else:
                return
        except oscore.ProtectionInvalid:
            if request.mtype == aiocoap.CON:
                raise error.BadRequest("Decryption failed")
            else:
                return

        unprotected.remote = OSCOREAddress(sc, request.remote)

        self.log.debug("Request %r was unprotected into %r", request, unprotected)

        sc = sc.context_for_response()

        inner_pipe = aiocoap.pipe.IterablePipe(unprotected)
        pr_that_can_take_errors = aiocoap.pipe.error_to_message(inner_pipe, self.log)
        # FIXME: do not create a task but run this in here (can this become a
        # feature of the aiterable PR?)
        aiocoap.pipe.run_driving_pipe(
                pr_that_can_take_errors,
                self._inner_site.render_to_pipe(inner_pipe),
                name="OSCORE response rendering for %r" % unprotected,
                )

        async for event in inner_pipe:
            if event.exception is not None:
                # These are expected to be rare in handlers
                #
                # FIXME should we try to render them? (See also
                # run_driving_pipe). Just raising them
                # would definitely be bad, as they might be renderable and
                # then would hit the outer message.
                self.log.warn("Turning error raised from renderer into nondescript protected error %r", event.exception)
                message = aiocoap.Message(code=aiocoap.INTERNAL_SERVER_ERROR)
                is_last = True
            else:
                message = event.message
                is_last = event.is_last

            # FIXME: Around several places in the use of pipe (and
            # now even here), non-final events are hard-coded as observations.
            # This should shift toward the source telling, or the stream being
            # annotated as "eventually consistent resource states".
            if not is_last:
                message.opt.observe = 0

            protected_response, _ = sc.protect(message, seqno)
            if message.opt.observe is not None:
                # FIXME: should be done in protect, or by something else that
                # generally handles obs numbers better (sending the
                # oscore-reconstructed number is nice because it's consistent
                # with a proxy that doesn't want to keep a counter when it
                # knows it's OSCORE already), but starting this per obs with
                # zero (unless it was done on that token recently) would be
                # most efficient
                protected_response.opt.observe = sc.sender_sequence_number & 0xffffffff
            self.log.debug("Response %r was encrypted into %r", message, protected_response)

            pipe.add_response(protected_response, is_last=is_last)
            if event.is_last:
                break
        # The created task gets cancelled here because the __aiter__ result is
        # dropped and thus all interest in the inner_pipe goes away



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module provides a slimmed-down replacement of the websockets_ module
(that regularly powers :mod:`aiocoap.transports.ws`) -- but implemented through
pyodide_'s JavaScript adapter towards the WebSocket module of the hosting
browser. It aims to be a drop-in replacement that provides the parts that can
be implemented in the browser, and to provide practical errors on the used
entry points. It will not go out of its way to mimick every aspect of the
websockets module, but restrain itself to what ``.ws`` needs.

**Future developement:** The module can probably be extended to cover all the
implementable functionality of websockets, and provide meaningful errors on all
its items. When that happens, it should be split out of aiocoap.

.. _websockets: https://websockets.readthedocs.io/
.. _pyodide: https://pyodide.org/
"""

import asyncio

import js
from pyodide.ffi.wrappers import add_event_listener

class WebSocketCommonProtocol:
    pass

class WebSocketClientProtocol(WebSocketCommonProtocol):
    def __init__(self, socket):
        # Note that this is an under-implemented constructor -- the real thing
        # is in `connect()` which is async enough to do more.

        self._socket = socket
        # FIXME: This is a workaround for WebSockets' shortcomings, while
        # WebSocketStreams are not deployed (see
        # https://developer.chrome.com/articles/websocketstream/ for details)

        self._queue = asyncio.Queue()

        # The initial setting doesn't matter too much because we're not handing
        # it out before setting this to True ... still feels cleaner this way.
        self.open = False

    async def recv(self):
        (etype, event) = await self._queue.get()
        if etype == "message":
            if isinstance(event.data, str):
                # FIXME: Test this
                return event.data
            return bytes((await event.data.arrayBuffer()).to_py())
        elif etype == "close":
            raise exceptions.ConnectionClosed()
        elif etype == "error":
            raise exceptions.WebSocketException("Connection error")
        else:
            raise RuntimeError("Unknown event in queue")

    async def send(self, msg):
        blob = js.Blob.new([js.Uint8Array.new(msg)])
        self._socket.send(blob)

    # FIXME: It'd be preferable if we could make this an unassigned property
    # that'd raise if anybody tried to access it (for neither do we know the
    # value, nor could anything useful be done with it), but as things are,
    # we'll have to rely on all users' sensibilities to not send around
    # addresses that are not globally usable. (The port, indicating the default
    # port, is an outright lie, though.)
    local_address = ("localhost", None)

    def on_message(self, event):
        self._queue.put_nowait(("message", event))

    def on_error(self, event):
        self.open = False
        self._queue.put_nowait(("error", event))

    def on_close(self, event):
        self.open = False
        self._queue.put_nowait(("close", event))

async def connect(uri, subprotocols=None, ping_interval=20, ssl=None) -> WebSocketClientProtocol:
    if ssl is not None:
        raise ValueError("SSL can not be configured within the browser WebSocket API")

    socket = js.WebSocket.new(uri, subprotocols)

    # Ignoring ping_interval: We can't tell what the browser does, and it may
    # be right nor not.

    proto = WebSocketClientProtocol(socket)

    add_event_listener(socket, "open", lambda e, q=proto._queue: q.put_nowait(("open", e)))
    add_event_listener(socket, "message", proto.on_message)
    add_event_listener(socket, "error", proto.on_error)
    add_event_listener(socket, "close", proto.on_close)
    (etype, event) = await proto._queue.get()
    if etype != "open":
        raise exceptions.WebSocketException("Failed to connect")
    proto.open = True

    return proto

class exceptions:
    """A class that is a good-enough approximation of ``websockets.exceptions``
    to get away with a single file implementing pyodide_websockets."""
    class WebSocketException(Exception):
        pass

    class ConnectionClosed(WebSocketException):
        pass

# Mocks required by the aiocoap.transports.ws module expecting a full implementation

class WebSocketServerProtocol:
    def __init__(self, *args, **kwargs):
        raise RuntimeError("Web sockets in web browsers can not be used as servers")

WebSocketServer = WebSocketServerProtocol

def serve(*args, **kwargs):
    WebSocketServer()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module contains the classes that are responsible for keeping track of
messages:

*   :class:`Context` roughly represents the CoAP endpoint (basically a UDP
    socket) -- something that can send requests and possibly can answer
    incoming requests.

*   a :class:`Request` gets generated whenever a request gets sent to keep
    track of the response

*   a :class:`Responder` keeps track of a single incoming request

Logging
~~~~~~~

Several constructors of the Context accept a logger name; these names go into
the construction of a Python logger.

Log events will be emitted to these on different levels, with "warning" and
above being a practical default for things that should may warrant reviewing by
an operator:

* DEBUG is used for things that occur even under perfect conditions.
* INFO is for things that are well expected, but might be interesting during
  testing a network of nodes and not just when debugging the library. (This
  includes timeouts, retransmissions, and pings.)
* WARNING is for everything that indicates a malbehaved peer. These don't
  *necessarily* indicate a client bug, though: Things like requesting a
  nonexistent block can just as well happen when a resource's content has
  changed between blocks. The library will not go out of its way to determine
  whether there is a plausible explanation for the odd behavior, and will
  report something as a warning in case of doubt.
* ERROR is used when something clearly went wrong. This includes irregular
  connection terminations and resource handler errors (which are demoted to
  error responses), and can often contain a backtrace.
"""

import asyncio
import weakref
import time
from typing import Optional, List

from . import defaults
from .credentials import CredentialsMap
from .message import Message
from .messagemanager import MessageManager
from .tokenmanager import TokenManager
from .pipe import Pipe, run_driving_pipe, error_to_message
from . import interfaces
from . import error
from .numbers import (INTERNAL_SERVER_ERROR, NOT_FOUND,
        CONTINUE, SHUTDOWN_TIMEOUT)
from .util.asyncio import py38args

import warnings
import logging


class Context(interfaces.RequestProvider):
    """Applications' entry point to the network

    A :class:`.Context` coordinates one or more network :mod:`.transports`
    implementations and dispatches data between them and the application.

    The application can start requests using the message dispatch methods, and
    set a :class:`resources.Site` that will answer requests directed to the
    application as a server.

    On the library-internals side, it is the prime implementation of the
    :class:`interfaces.RequestProvider` interface, creates :class:`Request` and
    :class:`Response` classes on demand, and decides which transport
    implementations to start and which are to handle which messages.

    **Context creation and destruction**

    The following functions are provided for creating and stopping a context:

    .. note::

        A typical application should only ever create one context, even (or
        especially when) it acts both as a server and as a client (in which
        case a server context should be created).

        A context that is not used any more must be shut down using
        :meth:`.shutdown()`, but typical applications will not need to because
        they use the context for the full process lifetime.

    .. automethod:: create_client_context
    .. automethod:: create_server_context

    .. automethod:: shutdown

    **Dispatching messages**

    CoAP requests can be sent using the following functions:

    .. automethod:: request

    If more control is needed, you can create a :class:`Request` yourself and
    pass the context to it.


    **Other methods and properties**

    The remaining methods and properties are to be considered unstable even
    when the project reaches a stable version number; please file a feature
    request for stabilization if you want to reliably access any of them.

    (Sorry for the duplicates, still looking for a way to make autodoc list
    everything not already mentioned).

    """
    def __init__(self, loop=None, serversite=None, loggername="coap", client_credentials=None, server_credentials=None):
        self.log = logging.getLogger(loggername)

        self.loop = loop or asyncio.get_event_loop()

        self.serversite = serversite

        self.request_interfaces = []

        self.client_credentials = client_credentials or CredentialsMap()
        self.server_credentials = server_credentials or CredentialsMap()

    #
    # convenience methods for class instanciation
    #

    async def _append_tokenmanaged_messagemanaged_transport(self, message_interface_constructor):
        tman = TokenManager(self)
        mman = MessageManager(tman)
        transport = await message_interface_constructor(mman)

        mman.message_interface = transport
        tman.token_interface = mman

        self.request_interfaces.append(tman)

    async def _append_tokenmanaged_transport(self, token_interface_constructor):
        tman = TokenManager(self)
        transport = await token_interface_constructor(tman)

        tman.token_interface = transport

        self.request_interfaces.append(tman)

    @classmethod
    async def create_client_context(cls, *, loggername="coap", loop=None, transports: Optional[List[str]] = None):
        """Create a context bound to all addresses on a random listening port.

        This is the easiest way to get a context suitable for sending client
        requests.
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        self = cls(loop=loop, serversite=None, loggername=loggername)

        selected_transports = transports or defaults.get_default_clienttransports(loop=loop)

        # FIXME make defaults overridable (postponed until they become configurable too)
        for transportname in selected_transports:
            if transportname == 'udp6':
                from .transports.udp6 import MessageInterfaceUDP6
                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceUDP6.create_client_transport_endpoint(mman, log=self.log, loop=loop))
            elif transportname == 'simple6':
                from .transports.simple6 import MessageInterfaceSimple6
                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceSimple6.create_client_transport_endpoint(mman, log=self.log, loop=loop))
            elif transportname == 'tinydtls':
                from .transports.tinydtls import MessageInterfaceTinyDTLS
                await self._append_tokenmanaged_messagemanaged_transport(

                    lambda mman: MessageInterfaceTinyDTLS.create_client_transport_endpoint(mman, log=self.log, loop=loop))
            elif transportname == 'tcpclient':
                from .transports.tcp import TCPClient
                await self._append_tokenmanaged_transport(
                    lambda tman: TCPClient.create_client_transport(tman, self.log, loop))
            elif transportname == 'tlsclient':
                from .transports.tls import TLSClient
                await self._append_tokenmanaged_transport(
                    lambda tman: TLSClient.create_client_transport(tman, self.log, loop, self.client_credentials))
            elif transportname == 'ws':
                from .transports.ws import WSPool
                await self._append_tokenmanaged_transport(
                    lambda tman: WSPool.create_transport(tman, self.log, loop, client_credentials=self.client_credentials))
            elif transportname == 'oscore':
                from .transports.oscore import TransportOSCORE
                oscoretransport = TransportOSCORE(self, self)
                self.request_interfaces.append(oscoretransport)
            else:
                raise RuntimeError("Transport %r not know for client context creation" % transportname)

        return self

    @classmethod
    async def create_server_context(cls, site, bind=None, *, loggername="coap-server", loop=None, _ssl_context=None, multicast=[], server_credentials=None, transports: Optional[List[str]] = None):
        """Create a context, bound to all addresses on the CoAP port (unless
        otherwise specified in the ``bind`` argument).

        This is the easiest way to get a context suitable both for sending
        client and accepting server requests.

        The ``bind`` argument, if given, needs to be a 2-tuple of IP address
        string and port number, where the port number can be None to use the default port.

        If ``multicast`` is given, it needs to be a list of (multicast address,
        interface name) tuples, which will all be joined. (The IPv4 style of
        selecting the interface by a local address is not supported; users may
        want to use the netifaces package to arrive at an interface name for an
        address).

        As a shortcut, the list may also contain interface names alone. Those
        will be joined for the 'all CoAP nodes' groups of IPv4 and IPv6 (with
        scopes 2 and 5) as well as the respective 'all nodes' groups in IPv6.

        Under some circumstances you may already need a context to pass into
        the site for creation; this is typically the case for servers that
        trigger requests on their own. For those cases, it is usually easiest
        to pass None in as a site, and set the fully constructed site later by
        assigning to the ``serversite`` attribute.
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        self = cls(loop=loop, serversite=site, loggername=loggername, server_credentials=server_credentials)

        multicast_done = not multicast

        selected_transports = transports or defaults.get_default_servertransports(loop=loop)

        for transportname in selected_transports:
            if transportname == 'udp6':
                from .transports.udp6 import MessageInterfaceUDP6

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceUDP6.create_server_transport_endpoint(mman, log=self.log, loop=loop, bind=bind, multicast=multicast))
                multicast_done = True
            # FIXME this is duplicated from the client version, as those are client-only anyway
            elif transportname == 'simple6':
                from .transports.simple6 import MessageInterfaceSimple6
                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceSimple6.create_client_transport_endpoint(mman, log=self.log, loop=loop))
            elif transportname == 'tinydtls':
                from .transports.tinydtls import MessageInterfaceTinyDTLS

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceTinyDTLS.create_client_transport_endpoint(mman, log=self.log, loop=loop))
            # FIXME end duplication
            elif transportname == 'tinydtls_server':
                from .transports.tinydtls_server import MessageInterfaceTinyDTLSServer

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceTinyDTLSServer.create_server(bind, mman, log=self.log, loop=loop, server_credentials=self.server_credentials))
            elif transportname == 'simplesocketserver':
                from .transports.simplesocketserver import MessageInterfaceSimpleServer
                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceSimpleServer.create_server(bind, mman, log=self.log, loop=loop))
            elif transportname == 'tcpserver':
                from .transports.tcp import TCPServer
                await self._append_tokenmanaged_transport(
                    lambda tman: TCPServer.create_server(bind, tman, self.log, loop))
            elif transportname == 'tcpclient':
                from .transports.tcp import TCPClient
                await self._append_tokenmanaged_transport(
                    lambda tman: TCPClient.create_client_transport(tman, self.log, loop))
            elif transportname == 'tlsserver':
                if _ssl_context is not None:
                    from .transports.tls import TLSServer
                    await self._append_tokenmanaged_transport(
                        lambda tman: TLSServer.create_server(bind, tman, self.log, loop, _ssl_context))
            elif transportname == 'tlsclient':
                from .transports.tls import TLSClient
                await self._append_tokenmanaged_transport(
                    lambda tman: TLSClient.create_client_transport(tman, self.log, loop, self.client_credentials))
            elif transportname == 'ws':
                from .transports.ws import WSPool
                await self._append_tokenmanaged_transport(
                    # None, None: Unlike the other transports this has a server/client generic creator, and only binds if there is some bind
                    lambda tman: WSPool.create_transport(tman, self.log, loop, client_credentials=self.client_credentials, server_bind=bind or (None, None), server_context=_ssl_context))
            elif transportname == 'oscore':
                from .transports.oscore import TransportOSCORE
                oscoretransport = TransportOSCORE(self, self)
                self.request_interfaces.append(oscoretransport)
            else:
                raise RuntimeError("Transport %r not know for server context creation" % transportname)

        if not multicast_done:
            self.log.warning("Multicast was requested, but no multicast capable transport was selected.")

        # This is used in tests to wait for externally launched servers to be ready
        self.log.debug("Server ready to receive requests")

        return self

    async def shutdown(self):
        """Take down any listening sockets and stop all related timers.

        After this coroutine terminates, and once all external references to
        the object are dropped, it should be garbage-collectable.

        This method takes up to
        :const:`aiocoap.numbers.constants.SHUTDOWN_TIMEOUT` seconds, allowing
        transports to perform any cleanup implemented in them (such as orderly
        connection shutdown and cancelling observations, where the latter is
        currently not implemented)."""

        self.log.debug("Shutting down context")

        done, pending = await asyncio.wait([
                asyncio.create_task(
                    ri.shutdown(),
                    **py38args(name="Shutdown of %r" % ri)
                    )
                for ri
                in self.request_interfaces],
            timeout=SHUTDOWN_TIMEOUT)
        for item in done:
            await item
        if pending:
            # Apart from being useful to see, this also ensures that developers
            # see the error in the logs during test suite runs -- and the error
            # should be easier to follow than the "we didn't garbage collect
            # everything" errors we see anyway (or otherwise, if the error is
            # escalated into a test failure)
            self.log.error("Shutdown timeout exceeded, returning anyway. Interfaces still busy: %s", pending)

    # FIXME: determine how official this should be, or which part of it is
    # public -- now that BlockwiseRequest uses it. (And formalize what can
    # change about messages and what can't after the remote has been thusly
    # populated).
    async def find_remote_and_interface(self, message):
        for ri in self.request_interfaces:
            if await ri.fill_or_recognize_remote(message):
                return ri
        raise RuntimeError("No request interface could route message")

    def request(self, request_message, handle_blockwise=True):
        if handle_blockwise:
            return BlockwiseRequest(self, request_message)

        pipe = Pipe(request_message, self.log)
        # Request sets up callbacks at creation
        result = Request(pipe, self.loop, self.log)

        async def send():
            try:
                request_interface = await self.find_remote_and_interface(request_message)
                request_interface.request(pipe)
            except Exception as e:
                pipe.add_exception(e)
                return
        self.loop.create_task(
                send(),
                **py38args(name="Request processing of %r" % result)
                )
        return result

    # the following are under consideration for moving into Site or something
    # mixed into it

    def render_to_pipe(self, pipe):
        """Fill a pipe by running the site's render_to_pipe interface and
        handling errors."""

        pr_that_can_receive_errors = error_to_message(pipe, self.log)

        run_driving_pipe(
                pr_that_can_receive_errors,
                self._render_to_pipe(pipe),
                name="Rendering for %r" % pipe.request,
                )

    async def _render_to_pipe(self, pipe):
        if self.serversite is None:
            pipe.add_response(Message(code=NOT_FOUND, payload=b"not a server"), is_last=True)
            return

        return await self.serversite.render_to_pipe(pipe)

class BaseRequest(object):
    """Common mechanisms of :class:`Request` and :class:`MulticastRequest`"""

class BaseUnicastRequest(BaseRequest):
    """A utility class that offers the :attr:`response_raising` and
    :attr:`response_nonraising` alternatives to waiting for the
    :attr:`response` future whose error states can be presented either as an
    unsuccessful response (eg. 4.04) or an exception.

    It also provides some internal tools for handling anything that has a
    :attr:`response` future and an :attr:`observation`"""

    @property
    async def response_raising(self):
        """An awaitable that returns if a response comes in and is successful,
        otherwise raises generic network exception or a
        :class:`.error.ResponseWrappingError` for unsuccessful responses.

        Experimental Interface."""

        response = await self.response
        if not response.code.is_successful():
            raise error.ResponseWrappingError(response)

        return response

    @property
    async def response_nonraising(self):
        """An awaitable that rather returns a 500ish fabricated message (as a
        proxy would return) instead of raising an exception.

        Experimental Interface."""

        # FIXME: Can we smuggle error_to_message into the underlying pipe?
        # That should make observe notifications into messages rather
        # than exceptions as well, plus it has fallbacks for `e.to_message()`
        # raising.

        try:
            return await self.response
        except error.RenderableError as e:
            return e.to_message()
        except Exception:
            return Message(code=INTERNAL_SERVER_ERROR)

class Request(interfaces.Request, BaseUnicastRequest):

    # FIXME: Implement timing out with REQUEST_TIMEOUT here

    def __init__(self, pipe, loop, log):
        self._pipe = pipe

        self.response = loop.create_future()

        if pipe.request.opt.observe == 0:
            self.observation = ClientObservation()
        else:
            self.observation = None

        self._runner = self._run()
        self._runner.send(None)
        def process(event):
            try:
                # would be great to have self or the runner as weak ref, but
                # see ClientObservation.register_callback comments -- while
                # that is around, we can't weakref here.
                self._runner.send(event)
                return True
            except StopIteration:
                return False
        self._stop_interest = self._pipe.on_event(process)

        self.log = log

        self.response.add_done_callback(self._response_cancellation_handler)

    def _response_cancellation_handler(self, response):
        # Propagate cancellation to the runner (if interest in the first
        # response is lost, there won't be observation items to pull out), but
        # not general completion (because if it's completed and not cancelled,
        # eg. when an observation is active)
        if self.response.cancelled() and self._runner is not None:
            # Dropping the only reference makes it stop with GeneratorExit,
            # similar to a cancelled task
            self._runner = None
            self._stop_interest()
        # Otherwise, there will be a runner still around, and it's its task to
        # call _stop_interest.

    @staticmethod
    def _add_response_properties(response, request):
        response.request = request

    def _run(self):
        # FIXME: This is in iterator form because it used to be a task that
        # awaited futures, and that code could be easily converted to an
        # iterator. I'm not sure that's a bad state here, but at least it
        # should be a more conscious decision to make this an iterator rather
        # than just having it happen to be one.
        #
        # FIXME: check that responses come from the same remmote as long as we're assuming unicast

        first_event = yield None

        if first_event.message is not None:
            self._add_response_properties(first_event.message, self._pipe.request)
            self.response.set_result(first_event.message)
        else:
            self.response.set_exception(first_event.exception)
            if not isinstance(first_event.exception, error.Error):
                self.log.warning(
                       "An exception that is not an aiocoap Error was raised "
                       "from a transport; please report this as a bug in "
                       "aiocoap: %r", first_event.exception)

        if self.observation is None:
            if not first_event.is_last:
                self.log.error("Pipe indicated more possible responses"
                               " while the Request handler would not know what to"
                               " do with them, stopping any further request.")
                self._stop_interest()
            return

        if first_event.is_last:
            self.observation.error(error.NotObservable())
            return

        if first_event.message.opt.observe is None:
            self.log.error("Pipe indicated more possible responses"
                           " while the Request handler would not know what to"
                           " do with them, stopping any further request.")
            self._stop_interest()
            return

        # variable names from RFC7641 Section 3.4
        v1 = first_event.message.opt.observe
        t1 = time.time()

        while True:
            # We don't really support cancellation of observations yet (see
            # https://github.com/chrysn/aiocoap/issues/92), but at least
            # stopping the interest is a way to free the local resources after
            # the first observation update, and to make the MID handler RST the
            # observation on the next.
            # FIXME: there *is* now a .on_cancel callback, we should at least
            # hook into that, and possibly even send a proper cancellation
            # then.
            next_event = yield True
            if self.observation.cancelled:
                self._stop_interest()
                return

            if next_event.exception is not None:
                self.observation.error(next_event.exception)
                if not next_event.is_last:
                    self._stop_interest()
                if not isinstance(next_event.exception, error.Error):
                    self.log.warning(
                           "An exception that is not an aiocoap Error was "
                           "raised from a transport during an observation; "
                           "please report this as a bug in aiocoap: %r",
                           next_event.exception)
                return

            self._add_response_properties(next_event.message, self._pipe.request)

            if next_event.message.opt.observe is not None:
                # check for reordering
                v2 = next_event.message.opt.observe
                t2 = time.time()

                is_recent = (v1 < v2 and v2 - v1 < 2**23) or \
                        (v1 > v2 and v1 - v2 > 2**23) or \
                        (t2 > t1 + self._pipe.request.transport_tuning.OBSERVATION_RESET_TIME)
                if is_recent:
                    t1 = t2
                    v1 = v2
            else:
                # the terminal message is always the last
                is_recent = True

            if is_recent:
                self.observation.callback(next_event.message)

            if next_event.is_last:
                self.observation.error(error.ObservationCancelled())
                return

            if next_event.message.opt.observe is None:
                self.observation.error(error.ObservationCancelled())
                self.log.error("Pipe indicated more possible responses"
                               " while the Request handler would not know what to"
                               " do with them, stopping any further request.")
                self._stop_interest()
                return


class BlockwiseRequest(BaseUnicastRequest, interfaces.Request):
    def __init__(self, protocol, app_request):
        self.protocol = protocol
        self.log = self.protocol.log.getChild("blockwise-requester")

        self.response = protocol.loop.create_future()

        if app_request.opt.observe is not None:
            self.observation = ClientObservation()
        else:
            self.observation = None

        self._runner = protocol.loop.create_task(self._run_outer(
            app_request,
            self.response,
            weakref.ref(self.observation) if self.observation is not None else lambda: None,
            self.protocol,
            self.log,
            ),
            **py38args(name="Blockwise runner for %r" % app_request))
        self.response.add_done_callback(self._response_cancellation_handler)

    def _response_cancellation_handler(self, response_future):
        # see Request._response_cancellation_handler
        if self.response.cancelled():
            self._runner.cancel()

    @classmethod
    async def _run_outer(cls, app_request, response, weak_observation, protocol, log):
        try:
            await cls._run(app_request, response, weak_observation, protocol, log)
        except asyncio.CancelledError:
            pass # results already set
        except Exception as e:
            logged = False
            if not response.done():
                logged = True
                response.set_exception(e)
            obs = weak_observation()
            if app_request.opt.observe is not None and obs is not None:
                logged = True
                obs.error(e)
            if not logged:
                # should be unreachable
                log.error("Exception in BlockwiseRequest runner neither went to response nor to observation: %s", e, exc_info=e)

    # This is a class method because that allows self and self.observation to
    # be freed even when this task is running, and the task to stop itself --
    # otherwise we couldn't know when users just "forget" about a request
    # object after using its response (esp. in observe cases) and leave this
    # task running.
    @classmethod
    async def _run(cls, app_request, response, weak_observation, protocol, log):
        # we need to populate the remote right away, because the choice of
        # blocks depends on it.
        await protocol.find_remote_and_interface(app_request)

        size_exp = app_request.remote.maximum_block_size_exp

        if app_request.opt.block1 is not None:
            assert app_request.opt.block1.block_number == 0, "Unexpected block number in app_request"
            assert not app_request.opt.block1.more, "Unexpected more-flag in app_request"
            # this is where the library user can traditionally pass in size
            # exponent hints into the library.
            size_exp = app_request.opt.block1.size_exponent

        # Offset in the message in blocks of size_exp. Whoever changes size_exp
        # is responsible for updating this number.
        block_cursor = 0

        while True:
            # ... send a chunk

            if size_exp >= 6:
                # FIXME from maximum_payload_size
                fragmentation_threshold = app_request.remote.maximum_payload_size
            else:
                fragmentation_threshold = (2 ** (size_exp + 4))

            if app_request.opt.block1 is not None or \
                    len(app_request.payload) > fragmentation_threshold:
                current_block1 = app_request._extract_block(
                        block_cursor,
                        size_exp,
                        app_request.remote.maximum_payload_size)
                if block_cursor == 0:
                    current_block1.opt.size1 = len(app_request.payload)
            else:
                current_block1 = app_request

            blockrequest = protocol.request(current_block1, handle_blockwise=False)
            blockresponse = await blockrequest.response

            # store for future blocks to ensure that the next blocks will be
            # sent from the same source address (in the UDP case; for many
            # other transports it won't matter).
            app_request.remote = blockresponse.remote

            if blockresponse.opt.block1 is None:
                if blockresponse.code.is_successful() and current_block1.opt.block1:
                    log.warning("Block1 option completely ignored by server, assuming it knows what it is doing.")
                # FIXME: handle 4.13 and retry with the indicated size option
                break

            block1 = blockresponse.opt.block1
            log.debug("Response with Block1 option received, number = %d, more = %d, size_exp = %d.", block1.block_number, block1.more, block1.size_exponent)

            if block1.block_number != current_block1.opt.block1.block_number:
                raise error.UnexpectedBlock1Option("Block number mismatch")

            if size_exp == 7:
                block_cursor += len(current_block1.payload) // 1024
            else:
                block_cursor += 1

            while block1.size_exponent < size_exp:
                block_cursor *= 2
                size_exp -= 1

            if not current_block1.opt.block1.more:
                if block1.more or blockresponse.code == CONTINUE:
                    # treating this as a protocol error -- letting it slip
                    # through would misrepresent the whole operation as an
                    # over-all 2.xx (successful) one.
                    raise error.UnexpectedBlock1Option("Server asked for more data at end of body")
                break

            # checks before preparing the next round:

            if blockresponse.opt.observe:
                # we're not *really* interested in that block, we just sent an
                # observe option to indicate that we'll want to observe the
                # resulting representation as a whole
                log.warning("Server answered Observe in early Block1 phase, cancelling the erroneous observation.")
                blockrequest.observe.cancel()

            if block1.more:
                # FIXME i think my own server is dowing this wrong
                #if response.code != CONTINUE:
                #    raise error.UnexpectedBlock1Option("more-flag set but no Continue")
                pass
            else:
                if not blockresponse.code.is_successful():
                    break
                else:
                    # ignoring (discarding) the successul intermediate result, waiting for a final one
                    continue

        lower_observation = None
        if app_request.opt.observe is not None:
            if blockresponse.opt.observe is not None:
                lower_observation = blockrequest.observation
            else:
                obs = weak_observation()
                if obs:
                    obs.error(error.NotObservable())
                del obs

        assert blockresponse is not None, "Block1 loop broke without setting a response"
        blockresponse.opt.block1 = None

        # FIXME check with RFC7959: it just says "send requests similar to the
        # requests in the Block1 phase", what does that mean? using the last
        # block1 as a reference for now, especially because in the
        # only-one-request-block case, that's the original request we must send
        # again and again anyway
        assembled_response = await cls._complete_by_requesting_block2(protocol, current_block1, blockresponse, log)

        response.set_result(assembled_response)
        # finally set the result

        if lower_observation is not None:
            # FIXME this can all be simplified a lot since it's no more
            # expected that observations shut themselves down when GC'd.
            obs = weak_observation()
            del weak_observation
            if obs is None:
                lower_observation.cancel()
                return
            future_weak_observation = protocol.loop.create_future() # packing this up because its destroy callback needs to reference the subtask
            subtask = asyncio.create_task(
                    cls._run_observation(
                        app_request,
                        lower_observation,
                        future_weak_observation,
                        protocol,
                        log),
                    **py38args(name="Blockwise observation for %r" % app_request)
                    )
            future_weak_observation.set_result(weakref.ref(obs, lambda obs: subtask.cancel()))
            obs.on_cancel(subtask.cancel)
            del obs
            await subtask

    @classmethod
    async def _run_observation(cls, original_request, lower_observation, future_weak_observation, protocol, log):
        weak_observation = await future_weak_observation
        # we can use weak_observation() here at any time, because whenever that
        # becomes None, this task gets cancelled
        try:
            async for block1_notification in lower_observation:
                log.debug("Notification received")
                full_notification = await cls._complete_by_requesting_block2(protocol, original_request, block1_notification, log)
                log.debug("Reporting completed notification")
                weak_observation().callback(full_notification)
            # FIXME verify that this loop actually ends iff the observation
            # was cancelled -- otherwise find out the cause(s) or make it not
            # cancel under indistinguishable circumstances
            weak_observation().error(error.ObservationCancelled())
        except asyncio.CancelledError:
            return
        except Exception as e:
            weak_observation().error(e)
        finally:
            # We generally avoid idempotent cancellation, but we may have
            # reached this point either due to an earlier cancellation or
            # without one
            if not lower_observation.cancelled:
                lower_observation.cancel()

    @classmethod
    async def _complete_by_requesting_block2(cls, protocol, request_to_repeat, initial_response, log):
        # FIXME this can probably be deduplicated against BlockwiseRequest

        if initial_response.opt.block2 is None or initial_response.opt.block2.more is False:
            initial_response.opt.block2 = None
            return initial_response

        if initial_response.opt.block2.block_number != 0:
            log.error("Error assembling blockwise response (expected first block)")
            raise error.UnexpectedBlock2()

        assembled_response = initial_response
        last_response = initial_response
        while True:
            current_block2 = request_to_repeat._generate_next_block2_request(assembled_response)

            current_block2 = current_block2.copy(remote=initial_response.remote)

            blockrequest = protocol.request(current_block2, handle_blockwise=False)
            last_response = await blockrequest.response

            if last_response.opt.block2 is None:
                log.warning("Server sent non-blockwise response after having started a blockwise transfer. Blockwise transfer cancelled, accepting single response.")
                return last_response

            block2 = last_response.opt.block2
            log.debug("Response with Block2 option received, number = %d, more = %d, size_exp = %d.", block2.block_number, block2.more, block2.size_exponent)
            try:
                assembled_response._append_response_block(last_response)
            except error.Error as e:
                log.error("Error assembling blockwise response, passing on error %r", e)
                raise

            if block2.more is False:
                return assembled_response

class ClientObservation:
    """An interface to observe notification updates arriving on a request.

    This class does not actually provide any of the observe functionality, it
    is purely a container for dispatching the messages via asynchronous
    iteration. It gets driven (ie. populated with responses or errors including
    observation termination) by a Request object.
    """
    def __init__(self):
        self.callbacks = []
        self.errbacks = []

        self.cancelled = False
        self._on_cancel = []

        self._latest_response = None
        # the analogous error is stored in _cancellation_reason when cancelled.

    def __aiter__(self):
        """`async for` interface to observations.

        This is the preferred interface to obtaining observations."""
        it = self._Iterator()
        self.register_callback(it.push, _suppress_deprecation=True)
        self.register_errback(it.push_err, _suppress_deprecation=True)
        return it

    class _Iterator:
        def __init__(self):
            self._future = asyncio.get_event_loop().create_future()

        def push(self, item):
            if self._future.done():
                # we don't care whether we overwrite anything, this is a lossy queue as observe is lossy
                self._future = asyncio.get_event_loop().create_future()
            self._future.set_result(item)

        def push_err(self, e):
            if self._future.done():
                self._future = asyncio.get_event_loop().create_future()
            self._future.set_exception(e)

        async def __anext__(self):
            f = self._future
            try:
                result = await self._future
                # FIXME see `await servobs._trigger` comment: might waiting for
                # the original future not yield the first future's result when
                # a quick second future comes in in a push?
                if f is self._future:
                    self._future = asyncio.get_event_loop().create_future()
                return result
            except (error.NotObservable, error.ObservationCancelled):
                # only exit cleanly when the server -- right away or later --
                # states that the resource is not observable any more
                # FIXME: check whether an unsuccessful message is still passed
                # as an observation result (or whether it should be)
                raise StopAsyncIteration

        def __del__(self):
            if self._future.done():
                try:
                    # Fetch the result so any errors show up at least in the
                    # finalizer output
                    self._future.result()
                except (error.ObservationCancelled, error.NotObservable):
                    # This is the case at the end of an observation cancelled
                    # by the server.
                    pass
                except error.NetworkError:
                    # This will already have shown up in the main result too.
                    pass
                except (error.LibraryShutdown, asyncio.CancelledError):
                    pass
                # Anything else flying out of this is unexpected and probably a
                # library error

    # When this function is removed, we can finally do cleanup better. Right
    # now, someone could register a callback that doesn't hold any references,
    # so we can't just stop the request when nobody holds a reference to this
    # any more. Once we're all in pull mode, we can make the `process` function
    # that sends data in here use a weak reference (because any possible
    # recipient would need to hold a reference to self or the iterator, and
    # thus _run).
    def register_callback(self, callback, _suppress_deprecation=False):
        """Call the callback whenever a response to the message comes in, and
        pass the response to it.

        The use of this function is deprecated: Use the asynchronous iteration
        interface instead."""
        if not _suppress_deprecation:
            warnings.warn("register_callback on observe results is deprected: Use `async for notify in request.observation` instead.",
                      DeprecationWarning, stacklevel=2)
        if self.cancelled:
            return

        self.callbacks.append(callback)
        if self._latest_response is not None:
            callback(self._latest_response)

    def register_errback(self, callback, _suppress_deprecation=False):
        """Call the callback whenever something goes wrong with the
        observation, and pass an exception to the callback. After such a
        callback is called, no more callbacks will be issued.

        The use of this function is deprecated: Use the asynchronous iteration
        interface instead."""
        if not _suppress_deprecation:
            warnings.warn("register_errback on observe results is deprected: Use `async for notify in request.observation` instead.",
                      DeprecationWarning, stacklevel=2)
        if self.cancelled:
            callback(self._cancellation_reason)
            return
        self.errbacks.append(callback)

    def callback(self, response):
        """Notify all listeners of an incoming response"""

        self._latest_response = response

        for c in self.callbacks:
            c(response)

    def error(self, exception):
        """Notify registered listeners that the observation went wrong. This
        can only be called once."""

        if self.errbacks is None:
            raise RuntimeError("Error raised in an already cancelled ClientObservation") from exception
        for c in self.errbacks:
            c(exception)

        self.cancel()
        self._cancellation_reason = exception

    def cancel(self):
        # FIXME determine whether this is called by anything other than error,
        # and make it private so there is always a _cancellation_reason
        """Cease to generate observation or error events. This will not
        generate an error by itself.

        This function is only needed while register_callback and
        register_errback are around; once their deprecations are acted on,
        dropping the asynchronous iterator will automatically cancel the
        observation.
        """

        assert not self.cancelled, "ClientObservation cancelled twice"

        # make sure things go wrong when someone tries to continue this
        self.errbacks = None
        self.callbacks = None

        self.cancelled = True
        while self._on_cancel:
            self._on_cancel.pop()()

        self._cancellation_reason = None

    def on_cancel(self, callback):
        if self.cancelled:
            callback()
        self._on_cancel.append(callback)

    def __repr__(self):
        return '<%s %s at %#x>' % (type(self).__name__, "(cancelled)" if self.cancelled else "(%s call-, %s errback(s))" % (len(self.callbacks), len(self.errbacks)), id(self))

class ServerObservation:
    def __init__(self):
        self._accepted = False
        self._trigger = asyncio.get_event_loop().create_future()
        # A deregistration is "early" if it happens before the response message
        # is actually sent; calling deregister() in that time (typically during
        # `render()`) will not send an unsuccessful response message but just
        # sent this flag which is set to None as soon as it is too late for an
        # early deregistration.
        # This mechanism is temporary until more of aiocoap behaves like
        # Pipe which does not suffer from this limitation.
        self._early_deregister = False
        self._late_deregister = False

    def accept(self, cancellation_callback):
        self._accepted = True
        self._cancellation_callback = cancellation_callback

    def deregister(self, reason=None):
        if self._early_deregister is False:
            self._early_deregister = True
            return

        warnings.warn("Late use of ServerObservation.deregister() is"
                      " deprecated, use .trigger with an unsuccessful value"
                      " instead",
                      DeprecationWarning)
        self.trigger(Message(code=INTERNAL_SERVER_ERROR, payload=b"Resource became unobservable"))

    def trigger(self, response=None, *, is_last=False):
        """Send an updated response; if None is given, the observed resource's
        rendering will be invoked to produce one.

        `is_last` can be set to True to indicate that no more responses will be
        sent. Note that an unsuccessful response will be the last no matter
        what is_last says, as such a message always terminates a CoAP
        observation."""
        if is_last:
            self._late_deregister = True
        if self._trigger.done():
            # we don't care whether we overwrite anything, this is a lossy queue as observe is lossy
            self._trigger = asyncio.get_event_loop().create_future()
        self._trigger.set_result(response)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""A pretty-printer for known mime types"""

import json
import re

import pygments
import pygments.lexers
import pygments.formatters

from aiocoap.util import linkformat, contenttype

from aiocoap.util.linkformat_pygments import _register

_register()

MEDIATYPE_HEXDUMP = 'text/vnd.aiocoap.hexdump'

def lexer_for_mime(mime):
    """A wrapper around pygments.lexers.get_lexer_for_mimetype that takes
    subtypes into consideration and catches the custom hexdump mime type."""

    if mime == MEDIATYPE_HEXDUMP:
        return pygments.lexers.HexdumpLexer()

    if mime == 'text/plain;charset=utf8':
        # We have fall-throughs in place anwyay, no need to go through a no-op
        # TextLexer
        raise pygments.util.ClassNotFound

    try:
        return pygments.lexers.get_lexer_for_mimetype(mime)
    except pygments.util.ClassNotFound:
        mime = re.sub('^([^/]+)/.*\\+([^;]+)(;.*)?$',
                lambda args: args[1] + '/' + args[2], mime)
        return pygments.lexers.get_lexer_for_mimetype(mime)

def pretty_print(message):
    """Given a CoAP message, reshape its payload into something human-readable.
    The return value is a triple (infos, mime, text) where text represents the
    payload, mime is a type that could be used to syntax-highlight the text
    (not necessarily related to the original mime type, eg. a report of some
    binary data that's shaped like Markdown could use a markdown mime type),
    and some line of infos that give additional data (like the reason for a hex
    dump or the original mime type).

    >>> from aiocoap import Message
    >>> def build(payload, request_cf, response_cf):
    ...     response = Message(payload=payload, content_format=response_cf)
    ...     request = Message(accept=request_cf)
    ...     response.request = request
    ...     return response
    >>> pretty_print(Message(payload=b"Hello", content_format=0))
    ([], 'text/plain;charset=utf8', 'Hello')
    >>> print(pretty_print(Message(payload=b'{"hello":"world"}', content_format=50))[-1])
    {
        "hello": "world"
    }
    >>> # Erroneous inputs still go to the pretty printer as long as they're
    >>> #Unicode
    >>> pretty_print(Message(payload=b'{"hello":"world', content_format=50))
    (['Invalid JSON not re-formated'], 'application/json', '{"hello":"world')
    >>> pretty_print(Message(payload=b'<>,', content_format=40))
    (['Invalid application/link-format content was not re-formatted'], 'application/link-format', '<>,')
    >>> pretty_print(Message(payload=b'a', content_format=60)) # doctest: +ELLIPSIS
    (['Showing hex dump of application/cbor payload: CBOR value is invalid'], 'text/vnd.aiocoap.hexdump', '00000000  61 ...
    """
    infos = []
    info = infos.append

    cf = message.opt.content_format or message.request.opt.accept
    if cf is None:
        content_type = "type unknown"
    elif cf.is_known():
        content_type = cf.media_type
        if cf.encoding != 'identity':
            info("Content format is %s in %s encoding; treating as "
                 "application/octet-stream because decompression is not "
                 "supported yet" % (cf.media_type, cf.encoding))
    else:
        content_type = "type %d" % cf
    category = contenttype.categorize(content_type)

    show_hex = None

    if linkformat is not None and category == 'link-format':
        try:
            decoded = message.payload.decode('utf8')
            try:
                parsed = linkformat.link_header.parse(decoded)
            except linkformat.link_header.ParseException:
                info("Invalid application/link-format content was not re-formatted")
                return (infos, 'application/link-format', decoded)
            else:
                info("application/link-format content was re-formatted")
                prettyprinted = ",\n".join(str(l) for l in parsed.links)
                return (infos, 'application/link-format', prettyprinted)
        except ValueError:
            # Handled later
            pass

    elif category in ('cbor', 'cbor-seq'):
        if category == 'cbor-seq':
            # Faking an indefinite length CBOR array is the easiest way to
            # parse an array into a list-like data structure, especially as
            # long as we don't indicate precise locations of invalid CBOR
            # anyway
            payload = b'\x9f' + message.payload + b'\xff'
        else:
            payload = message.payload

        try:
            import cbor_diag

            formatted = cbor_diag.cbor2diag(payload)

            if category == 'cbor-seq':
                info("CBOR sequence message shown as array in Diagnostic Notation")
            else:
                info("CBOR message shown in Diagnostic Notation")

            # It's not exactly CDDL, but it's close enough that the syntax
            # highlighting looks OK, and tolerant enough to not complain about
            # missing leading barewords and "=" signs
            return (infos, 'text/x-cddl', formatted)
        except ImportError:
            show_hex = "No CBOR pretty-printer available"
        except ValueError:
            show_hex = "CBOR value is invalid"

    elif category == 'json':
        try:
            decoded = message.payload.decode('utf8')
        except ValueError:
            pass
        else:
            try:
                parsed = json.loads(decoded)
            except ValueError:
                info("Invalid JSON not re-formated")
                return (infos, 'application/json', decoded)
            else:
                info("JSON re-formated and indented")
                formatted = json.dumps(parsed, indent=4)
                return (infos, 'application/json', formatted)

    # That's about the formats we do for now.

    if show_hex is None:
        try:
            text = message.payload.decode('utf8')
        except UnicodeDecodeError:
            show_hex = "Message can not be parsed as UTF-8"
        else:
            return (infos, 'text/plain;charset=utf8', text)

    info("Showing hex dump of %s payload%s" % (
        content_type if cf is not None else "untyped",
        ": " + show_hex if show_hex is not None else ""))
    data = message.payload
    # Not the most efficient hex dumper, but we won't stream video over
    # this anyway
    formatted = []
    offset = 0
    while data:
        line, data = data[:16], data[16:]

        formatted.append("%08x  " % offset +
                " ".join("%02x" % line[i] if i < len(line) else "  " for i in range(8)) + "  " +
                " ".join("%02x" % line[i] if i < len(line) else "  " for i in range(8, 16)) + "  |" +
                "".join(chr(x) if 32 <= x < 127 else '.' for x in line) +
                "|\n")

        offset += len(line)
    if offset % 16 != 0:
        formatted.append("%08x\n" % offset)
    return (infos, MEDIATYPE_HEXDUMP, "".join(formatted))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from itertools import chain

from .numbers.optionnumbers import OptionNumber
from .error import UnparsableMessage

def _read_extended_field_value(value, rawdata):
    """Used to decode large values of option delta and option length
       from raw binary form."""
    if value >= 0 and value < 13:
        return (value, rawdata)
    elif value == 13:
        if len(rawdata) < 1:
            raise UnparsableMessage("Option ended prematurely")
        return (rawdata[0] + 13, rawdata[1:])
    elif value == 14:
        if len(rawdata) < 2:
            raise UnparsableMessage("Option ended prematurely")
        return (int.from_bytes(rawdata[:2], 'big') + 269, rawdata[2:])
    else:
        raise UnparsableMessage("Option contained partial payload marker.")


def _write_extended_field_value(value):
    """Used to encode large values of option delta and option length
       into raw binary form.
       In CoAP option delta and length can be represented by a variable
       number of bytes depending on the value."""
    if value >= 0 and value < 13:
        return (value, b'')
    elif value >= 13 and value < 269:
        return (13, (value - 13).to_bytes(1, 'big'))
    elif value >= 269 and value < 65804:
        return (14, (value - 269).to_bytes(2, 'big'))
    else:
        raise ValueError("Value out of range.")


def _single_value_view(option_number, doc=None):
    """Generate a property for a given option number, where the option is not
    repeatable. For getting, it will return the value of the first option
    object with matching number. For setting, it will remove all options with
    that number and create one with the given value. The property can be
    deleted, resulting in removal of the option from the header.

    For consistency, setting the value to None also clears the option. (Note
    that with the currently implemented optiontypes, None is not a valid value
    for any of them)."""

    def _getter(self, option_number=option_number):
        options = self.get_option(option_number)
        if not options:
            return None
        else:
            return options[0].value

    def _setter(self, value, option_number=option_number):
        self.delete_option(option_number)
        if value is not None:
            self.add_option(option_number.create_option(value=value))

    def _deleter(self, option_number=option_number):
        self.delete_option(option_number)

    return property(_getter, _setter, _deleter, doc or "Single-value view on the %s option." % option_number)

def _items_view(option_number, doc=None):
    """Generate a property for a given option number, where the option is
    repeatable. For getting, it will return a tuple of the values of the option
    objects with matching number. For setting, it will remove all options with
    that number and create new ones from the given iterable."""

    def _getter(self, option_number=option_number):
        return tuple(o.value for o in self.get_option(option_number))

    def _setter(self, value, option_number=option_number):
        self.delete_option(option_number)
        for v in value:
            self.add_option(option_number.create_option(value=v))

    def _deleter(self, option_number=option_number):
        self.delete_option(option_number)

    return property(_getter, _setter, _deleter, doc=doc or "Iterable view on the %s option." % option_number)

def _empty_presence_view(option_number, doc=None):
    """Generate a property for a given option number, where the option is not
    repeatable and (usually) empty. The values True and False are mapped to
    presence and absence of the option."""

    def _getter(self, option_number=option_number):
        return bool(self.get_option(option_number))

    def _setter(self, value, option_number=option_number):
        self.delete_option(option_number)
        if value:
            self.add_option(option_number.create_option())

    return property(_getter, _setter, doc=doc or "Presence of the %s option." % option_number)

class Options(object):
    """Represent CoAP Header Options."""

    # this is not so much an optimization as a safeguard -- if custom
    # attributes were placed here, they could be accessed but would not be
    # serialized
    __slots__ = ["_options"]

    def __init__(self):
        self._options = {}

    def __eq__(self, other):
        if not isinstance(other, Options):
            return NotImplemented
        # this implementation is much easier than implementing equality on
        # StringOption etc
        return self.encode() == other.encode()

    def __repr__(self):
        text = ", ".join("%s: %s" % (OptionNumber(k), " / ".join(map(str, v))) for (k, v) in self._options.items())
        return "<aiocoap.options.Options at %#x: %s>" % (id(self), text or "empty")

    def _repr_html_(self):
        if self._options:
            n_opt = sum(len(o) for o in self._options.values())
            items = (f'<li value="{int(k)}">{OptionNumber(k)._repr_html_()}: {", ".join(vi._repr_html_() for vi in v)}' for (k, v) in sorted(self._options.items()))
            return f"""<details><summary style="display:list-item">{n_opt} option{'s' if n_opt != 1 else ''}</summary><ol>{''.join(items)}</ol></details>"""
        else:
            return "<div>No options</div>"

    def decode(self, rawdata):
        """Passed a CoAP message body after the token as rawdata, fill self
        with the options starting at the beginning of rawdata, an return the
        rest of the message (the body)."""
        option_number = OptionNumber(0)

        while rawdata:
            if rawdata[0] == 0xFF:
                return rawdata[1:]
            dllen = rawdata[0]
            delta = (dllen & 0xF0) >> 4
            length = (dllen & 0x0F)
            rawdata = rawdata[1:]
            (delta, rawdata) = _read_extended_field_value(delta, rawdata)
            (length, rawdata) = _read_extended_field_value(length, rawdata)
            option_number += delta
            if len(rawdata) < length:
                raise UnparsableMessage("Option announced but absent")
            option = option_number.create_option(decode=rawdata[:length])
            self.add_option(option)
            rawdata = rawdata[length:]
        return b''

    def encode(self):
        """Encode all options in option header into string of bytes."""
        data = []
        current_opt_num = 0
        for option in self.option_list():
            optiondata = option.encode()

            delta, extended_delta = _write_extended_field_value(option.number - current_opt_num)
            length, extended_length = _write_extended_field_value(len(optiondata))

            data.append(bytes([((delta & 0x0F) << 4) + (length & 0x0F)]))
            data.append(extended_delta)
            data.append(extended_length)
            data.append(optiondata)

            current_opt_num = option.number

        return b''.join(data)

    def add_option(self, option):
        """Add option into option header."""
        self._options.setdefault(option.number, []).append(option)

    def delete_option(self, number):
        """Delete option from option header."""
        if number in self._options:
            self._options.pop(number)

    def get_option(self, number):
        """Get option with specified number."""
        return self._options.get(number, ())

    def option_list(self):
        return chain.from_iterable(sorted(self._options.values(), key=lambda x: x[0].number))

    uri_path = _items_view(OptionNumber.URI_PATH)
    uri_query = _items_view(OptionNumber.URI_QUERY)
    location_path = _items_view(OptionNumber.LOCATION_PATH)
    location_query = _items_view(OptionNumber.LOCATION_QUERY)
    block2 = _single_value_view(OptionNumber.BLOCK2)
    block1 = _single_value_view(OptionNumber.BLOCK1)
    content_format = _single_value_view(OptionNumber.CONTENT_FORMAT)
    etag = _single_value_view(OptionNumber.ETAG, "Single ETag as used in responses")
    etags = _items_view(OptionNumber.ETAG, "List of ETags as used in requests")
    if_none_match = _empty_presence_view(OptionNumber.IF_NONE_MATCH)
    observe = _single_value_view(OptionNumber.OBSERVE)
    accept = _single_value_view(OptionNumber.ACCEPT)
    uri_host = _single_value_view(OptionNumber.URI_HOST)
    uri_port = _single_value_view(OptionNumber.URI_PORT)
    proxy_uri = _single_value_view(OptionNumber.PROXY_URI)
    proxy_scheme = _single_value_view(OptionNumber.PROXY_SCHEME)
    size1 = _single_value_view(OptionNumber.SIZE1)
    object_security = _single_value_view(OptionNumber.OBJECT_SECURITY)
    max_age = _single_value_view(OptionNumber.MAX_AGE)
    if_match = _items_view(OptionNumber.IF_MATCH)
    no_response = _single_value_view(OptionNumber.NO_RESPONSE)
    echo = _single_value_view(OptionNumber.ECHO)
    request_tag = _items_view(OptionNumber.REQUEST_TAG)
    hop_limit = _single_value_view(OptionNumber.HOP_LIMIT)
    request_hash = _single_value_view(OptionNumber.REQUEST_HASH,
        "Experimental property for draft-amsuess-core-cachable-oscore")
    edhoc = _empty_presence_view(OptionNumber.EDHOC)
    size2 = _single_value_view(OptionNumber.SIZE2)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

#: Make library version internally
#:
#: This is not supposed to be used in any decision-making process (use package
#: dependencies for that) or workarounds, but used by command-line tools or the
#: impl-info link to provide debugging information.
version = "0.4.7.post0"

#: URI used to describe the current version of the library
#:
#: This is used the same way as `version` but when a URI is required, for
#: example as a default value for .well-known/core's rel=impl-info link.
library_uri = "https://christian.amsuess.com/tools/aiocoap/#version-" + version



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio
import socket

from aiocoap.transports import rfc8323common
from aiocoap import interfaces, error, util
from aiocoap import COAP_PORT, Message
from aiocoap import defaults
from ..util.asyncio import py38args

def _extract_message_size(data: bytes):
    """Read out the full length of a CoAP messsage represented by data.

    Returns None if data is too short to read the (full) length.

    The number returned is the number of bytes that has to be read into data to
    start reading the next message; it consists of a constant term, the token
    length and the extended length of options-plus-payload."""

    if not data:
        return None

    l = data[0] >> 4
    tokenoffset = 2
    tkl = data[0] & 0x0f

    if l >= 13:
        if l == 13:
            extlen = 1
            offset = 13
        elif l == 14:
            extlen = 2
            offset = 269
        else:
            extlen = 4
            offset = 65805
        if len(data) < extlen + 1:
            return None
        tokenoffset = 2 + extlen
        l = int.from_bytes(data[1:1 + extlen], "big") + offset
    return tokenoffset, tkl, l

def _decode_message(data: bytes) -> Message:
    tokenoffset, tkl, _ = _extract_message_size(data)
    if tkl > 8:
        raise error.UnparsableMessage("Overly long token")
    code = data[tokenoffset - 1]
    token = data[tokenoffset:tokenoffset + tkl]

    msg = Message(code=code, token=token)

    msg.payload = msg.opt.decode(data[tokenoffset + tkl:])

    return msg

def _encode_length(l: int):
    if l < 13:
        return (l, b"")
    elif l < 269:
        return (13, (l - 13).to_bytes(1, 'big'))
    elif l < 65805:
        return (14, (l - 269).to_bytes(2, 'big'))
    else:
        return (15, (l - 65805).to_bytes(4, 'big'))

def _serialize(msg: Message) -> bytes:
    data = [msg.opt.encode()]
    if msg.payload:
        data += [b'\xff', msg.payload]
    data = b"".join(data)
    l, extlen = _encode_length(len(data))

    tkl = len(msg.token)
    if tkl > 8:
        raise ValueError("Overly long token")

    return b"".join((
            bytes(((l << 4) | tkl,)),
            extlen,
            bytes((msg.code,)),
            msg.token,
            data
            ))

class TcpConnection(asyncio.Protocol, rfc8323common.RFC8323Remote, interfaces.EndpointAddress):
    # currently, both the protocol and the EndpointAddress are the same object.
    # if, at a later point in time, the keepaliving of TCP connections should
    # depend on whether the library user still keeps a usable address around,
    # those functions could be split.

    def __init__(self, ctx, log, loop, *, is_server):
        super().__init__()
        self._ctx = ctx
        self.log = log
        self.loop = loop

        self._spool = b""

        self._remote_settings = None

        self._transport = None
        self._local_is_server = is_server

    @property
    def scheme(self):
        return self._ctx._scheme

    def _send_message(self, msg: Message):
        self.log.debug("Sending message: %r", msg)
        self._transport.write(_serialize(msg))

    def _abort_with(self, abort_msg):
        if self._transport is not None:
            self._send_message(abort_msg)
            self._transport.close()
        else:
            # FIXME: find out how this happens; i've only seen it after nmap
            # runs against an aiocoap server and then shutting it down.
            # "poisoning" the object to make sure this can not be exploited to
            # bypass the server shutdown.
            self._ctx = None

    # implementing asyncio.Protocol

    def connection_made(self, transport):
        self._transport = transport

        ssl_object = transport.get_extra_info('ssl_object')
        if ssl_object is not None:
            server_name = getattr(ssl_object, "indicated_server_name", None)
        else:
            server_name = None

        # `host` already contains the interface identifier, so throwing away
        # scope and interface identifier
        self._local_hostinfo = transport.get_extra_info('sockname')[:2]
        self._remote_hostinfo = transport.get_extra_info('peername')[:2]

        def none_default_port(sockname):
            return (sockname[0], None if sockname[1] == self._ctx._default_port else sockname[1])
        self._local_hostinfo = none_default_port(self._local_hostinfo)
        self._remote_hostinfo = none_default_port(self._remote_hostinfo)

        # SNI information available
        if server_name is not None:
            if self._local_is_server:
                self._local_hostinfo = (server_name, self._local_hostinfo[1])
            else:
                self._remote_hostinfo = (server_name, self._remote_hostinfo[1])

        self._send_initial_csm()

    def connection_lost(self, exc):
        # FIXME react meaningfully:
        # * send event through pool so it can propagate the error to all
        #   requests on the same remote
        # * mark the address as erroneous so it won't be recognized by
        #   fill_or_recognize_remote

        self._ctx._dispatch_error(self, exc)

    def data_received(self, data):
        # A rope would be more efficient here, but the expected case is that
        # _spool is b"" and spool gets emptied soon -- most messages will just
        # fit in a single TCP package and not be nagled together.
        #
        # (If this does become a bottleneck, say self._spool = SomeRope(b"")
        # and barely change anything else).

        self._spool += data

        while True:
            msglen = _extract_message_size(self._spool)
            if msglen is None:
                break
            msglen = sum(msglen)
            if msglen > self._my_max_message_size:
                self.abort("Overly large message announced")
                return

            if msglen > len(self._spool):
                break

            msg = self._spool[:msglen]
            try:
                msg = _decode_message(msg)
            except error.UnparsableMessage:
                self.abort("Failed to parse message")
                return
            msg.remote = self

            self.log.debug("Received message: %r", msg)

            self._spool = self._spool[msglen:]

            if msg.code.is_signalling():
                try:
                    self._process_signaling(msg)
                except rfc8323common.CloseConnection as e:
                    self._ctx._dispatch_error(self, e.args[0])
                    self._transport.close()
                continue

            if self._remote_settings is None:
                self.abort("No CSM received")
                return

            self._ctx._dispatch_incoming(self, msg)

    def eof_received(self):
        # FIXME: as with connection_lost, but less noisy if announced
        # FIXME: return true and initiate own shutdown if that is what CoAP prescribes
        pass

    def pause_writing(self):
        # FIXME: do something ;-)
        pass

    def resume_writing(self):
        # FIXME: do something ;-)
        pass

    # RFC8323Remote.release recommends subclassing this, but there's no easy
    # awaitable here yet, and no important business to finish, timeout-wise.

class _TCPPooling:
    # implementing TokenInterface

    def send_message(self, message, messageerror_monitor):
        # Ignoring messageerror_monitor: CoAP over reliable transports has no
        # way of indicating that a particular message was bad, it always shuts
        # down the complete connection

        if message.code.is_response():
            no_response = (message.opt.no_response or 0) & (1 << message.code.class_ - 1) != 0
            if no_response:
                return

        message.opt.no_response = None

        message.remote._send_message(message)

    # used by the TcpConnection instances

    def _dispatch_incoming(self, connection, msg):
        if msg.code == 0:
            pass

        if msg.code.is_response():
            self._tokenmanager.process_response(msg)
            # ignoring the return value; unexpected responses can be the
            # asynchronous result of cancelled observations
        else:
            self._tokenmanager.process_request(msg)

    def _dispatch_error(self, connection, exc):
        self._evict_from_pool(connection)

        if self._tokenmanager is None:
            if exc is not None:
                self.log.warning("Ignoring late error during shutdown: %s", exc)
            else:
                # it's just a regular connection loss, that's to be expected during shutdown
                pass
            return

        self._tokenmanager.dispatch_error(exc, connection)

    # for diverting behavior of _TLSMixIn
    _scheme = 'coap+tcp'
    _default_port = COAP_PORT

class TCPServer(_TCPPooling, interfaces.TokenInterface):
    def __init__(self):
        self._pool = set()

    @classmethod
    async def create_server(cls, bind, tman: interfaces.TokenManager, log, loop, *, _server_context=None):
        self = cls()
        self._tokenmanager = tman
        self.log = log
        #self.loop = loop

        bind = bind or ('::', None)
        bind = (bind[0], bind[1] + (self._default_port - COAP_PORT) if bind[1] else self._default_port)

        def new_connection():
            c = TcpConnection(self, log, loop, is_server=True)
            self._pool.add(c)
            return c

        try:
            server = await loop.create_server(new_connection, bind[0], bind[1],
                    ssl=_server_context, reuse_port=defaults.has_reuse_port())
        except socket.gaierror as e:
            raise error.ResolutionError("No local bindable address found for %s" % bind[0]) from e
        self.server = server

        return self

    def _evict_from_pool(self, connection):
        # May easily happen twice, once when an error comes in and once when
        # the connection is (subsequently) closed.
        if connection in self._pool:
            self._pool.remove(connection)

    # implementing TokenInterface

    async def fill_or_recognize_remote(self, message):
        if message.remote is not None \
                and isinstance(message.remote, TcpConnection) \
                and message.remote._ctx is self:
            return True

        return False

    async def shutdown(self):
        self.log.debug("Shutting down server %r", self)
        self._tokenmanager = None
        self.server.close()
        # Since server has been closed, we won't be getting any *more*
        # connections, so we can process them all now:
        shutdowns = [
                asyncio.create_task(
                    c.release(),
                    **py38args(name="Close client %s" % c))
                for c
                in self._pool
                ]
        shutdowns.append(asyncio.create_task(
                self.server.wait_closed(),
                **py38args(name="Close server %s" % self)))
        # There is at least one member, so we can just .wait()
        await asyncio.wait(shutdowns)

class TCPClient(_TCPPooling, interfaces.TokenInterface):
    def __init__(self):
        self._pool = {} # (host, port) -> connection
        # note that connections are filed by host name, so different names for
        # the same address might end up with different connections, which is
        # probably okay for TCP, and crucial for later work with TLS.

    async def _spawn_protocol(self, message):
        if message.unresolved_remote is None:
            host = message.opt.uri_host
            port = message.opt.uri_port or self._default_port
            if host is None:
                raise ValueError("No location found to send message to (neither in .opt.uri_host nor in .remote)")
        else:
            host, port = util.hostportsplit(message.unresolved_remote)
            port = port or self._default_port

        if (host, port) in self._pool:
            return self._pool[(host, port)]

        try:
            _, protocol = await self.loop.create_connection(
                    lambda: TcpConnection(self, self.log, self.loop,
                        is_server=False),
                    host, port,
                    ssl=self._ssl_context_factory(message.unresolved_remote))
        except socket.gaierror as e:
            raise error.ResolutionError("No address information found for requests to %r" % host) from e
        except OSError as e:
            raise error.NetworkError("Connection failed to %r" % host) from e

        self._pool[(host, port)] = protocol

        return protocol

    # for diverting behavior of TLSClient
    def _ssl_context_factory(self, hostinfo):
        return None

    def _evict_from_pool(self, connection):
        keys = []
        for k, p in self._pool.items():
            if p is connection:
                keys.append(k)
        # should really be zero or one
        for k in keys:
            self._pool.pop(k)

    @classmethod
    async def create_client_transport(cls, tman: interfaces.TokenManager, log, loop, credentials=None):
        # this is not actually asynchronous, and even though the interface
        # between the context and the creation of interfaces is not fully
        # standardized, this stays in the other inferfaces' style.
        self = cls()
        self._tokenmanager = tman
        self.log = log
        self.loop = loop
        # used by the TLS variant; FIXME not well thought through
        self.credentials = credentials

        return self

    # implementing TokenInterface

    async def fill_or_recognize_remote(self, message):
        if message.remote is not None \
                and isinstance(message.remote, TcpConnection) \
                and message.remote._ctx is self:
            return True

        if message.requested_scheme == self._scheme:
            # FIXME: This could pool outgoing connections.
            # (Checking if an incoming connection is a pool candidate is
            # probably overkill because even if a URI can be constructed from a
            # ephemeral client port, nobody but us can use it, and we can just
            # set .remote).
            message.remote = await self._spawn_protocol(message)
            return True

        return False

    async def shutdown(self):
        self.log.debug("Shutting down any outgoing connections on on %r", self)
        self._tokenmanager = None

        shutdowns = [asyncio.create_task(
            c.release(),
            **py38args(name="Close client %s" % c))
            for c in self._pool.values()]
        if not shutdowns:
            # wait is documented to require a non-empty set
            return
        await asyncio.wait(shutdowns)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from aiocoap import interfaces, error, util
from aiocoap import COAP_PORT, Message

class GenericMessageInterface(interfaces.MessageInterface):
    """GenericMessageInterface is not a standalone implementation of a
    message inteface. It does implement everything between the MessageInterface
    and a not yet fully specified interface of "bound UDP sockets".

    It delegates sending through the address objects (which persist through
    some time, given this is some kind of bound-socket scenario).

    The user must:
    * set up a ._pool after construction with a shutdown and a connect method
    * provide their addresses with a send(bytes) method
    * pass incoming data to the _received_datagram and _received_exception methods
    """

    def __init__(self, mman: interfaces.MessageManager, log, loop):
        self._mman = mman
        self._log = log
        self._loop = loop

    # Callbacks to be hooked up by the user of the class; feed data on to the
    # message manager

    def _received_datagram(self, address, datagram):
        try:
            message = Message.decode(datagram, remote=address)
        except error.UnparsableMessage:
            self._log.warning("Ignoring unparsable message from %s", address)
            return

        self._mman.dispatch_message(message)

    def _received_exception(self, address, exception):
        self._mman.dispatch_error(exception, address)

    # Implementations of MessageInterface

    def send(self, message):
        if self._mman is None:
            self._log.info("Not sending message %r: transport is already shutting down.", message)
        else:
            message.remote.send(message.encode())

    async def shutdown(self):
        await self._pool.shutdown()
        self._mman = None

    async def determine_remote(self, request):
        if request.requested_scheme not in ('coap', None):
            return None

        if request.unresolved_remote is not None:
            host, port = util.hostportsplit(request.unresolved_remote)
            port = port or COAP_PORT
        elif request.opt.uri_host:
            host = request.opt.uri_host
            port = request.opt.uri_port or COAP_PORT
        else:
            raise ValueError("No location found to send message to (neither in .opt.uri_host nor in .remote)")

        return await self._pool.connect((host, port))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module contains helpers that inspect available modules and platform
specifics to give sane values to aiocoap defaults.

All of this should eventually overridable by other libraries wrapping/using
aiocoap and by applications using aiocoap; however, these overrides do not
happen in the defaults module but where these values are actually accessed, so
this module is considered internal to aiocoap and not part of the API.

The ``_missing_modules`` functions are helpers for inspecting what is
reasonable to expect to work. They can influence default values, but should not
be used in the rest of the code for feature checking (just raise the
ImportErrors) unless it's directly user-visible ("You configured OSCORE key
material, but OSCORE needs the following unavailable modules") or in the test
suite to decide which tests to skip.
"""

import os
import socket
import sys
import warnings

try:
    import pyodide  # noqa: F401
    import js       # noqa: F401
except ImportError:
    is_pyodide = False
else:
    is_pyodide = True

def get_default_clienttransports(*, loop=None, use_env=True):
    """Return a list of transports that should be connected when a client
    context is created.

    If an explicit ``AIOCOAP_CLIENT_TRANSPORT`` environment variable is set, it
    is read as a colon separated list of transport names.

    By default, a DTLS mechanism will be picked if the required modules are
    available, and a UDP transport will be selected depending on whether the
    full udp6 transport is known to work.
    """

    if use_env and 'AIOCOAP_CLIENT_TRANSPORT' in os.environ:
        yield from os.environ['AIOCOAP_CLIENT_TRANSPORT'].split(':')
        return

    if not oscore_missing_modules():
        yield 'oscore'

    if not dtls_missing_modules():
        yield 'tinydtls'

    yield 'tcpclient'
    yield 'tlsclient'
    if not ws_missing_modules():
        yield 'ws'

    if sys.platform != 'linux':
        # udp6 was never reported to work on anything but linux; would happily
        # add more platforms.
        yield 'simple6'
        return

    # on android it seems that it's only the AI_V4MAPPED that causes trouble,
    # that should be managable in udp6 too.
    yield 'udp6'
    return

def get_default_servertransports(*, loop=None, use_env=True):
    """Return a list of transports that should be connected when a server
    context is created.

    If an explicit ``AIOCOAP_SERVER_TRANSPORT`` environment variable is set, it
    is read as a colon separated list of transport names.

    By default, a DTLS mechanism will be picked if the required modules are
    available, and a UDP transport will be selected depending on whether the
    full udp6 transport is known to work. Both a simple6 and a simplesocketserver
    will be selected when udp6 is not available, and the simple6 will be used
    for any outgoing requests, which the simplesocketserver could serve but is worse
    at.
    """

    if use_env and 'AIOCOAP_SERVER_TRANSPORT' in os.environ:
        yield from os.environ['AIOCOAP_SERVER_TRANSPORT'].split(':')
        return

    if not oscore_missing_modules():
        yield 'oscore'

    if not dtls_missing_modules():
        if 'AIOCOAP_DTLSSERVER_ENABLED' in os.environ:
            yield 'tinydtls_server'
        yield 'tinydtls'

    yield 'tcpserver'
    yield 'tcpclient'
    yield 'tlsserver'
    yield 'tlsclient'
    if not ws_missing_modules():
        yield 'ws'

    if sys.platform != 'linux':
        # udp6 was never reported to work on anything but linux; would happily
        # add more platforms.
        yield 'simple6'
        yield 'simplesocketserver'
        return

    # on android it seems that it's only the AI_V4MAPPED that causes trouble,
    # that should be managable in udp6 too.
    yield 'udp6'
    return

def has_reuse_port(*, use_env=True):
    """Return true if the platform indicates support for SO_REUSEPORT.

    Can be overridden by explicitly setting ``AIOCOAP_REUSE_PORT`` to 1 or
    0."""

    if use_env and os.environ.get('AIOCOAP_REUSE_PORT'):
        return bool(int(os.environ['AIOCOAP_REUSE_PORT']))

    return hasattr(socket, 'SO_REUSEPORT')

def use_ai_v4mapped_emulation():
    """This used to indicate when ai_v4mapped emulation was used. Given it is
    not used at all any more, the function is deprecated."""
    warnings.warn("AI_V4MAPPED emulation is not used any more at all",
            warnings.DeprecationWarning)
    return False

# FIXME: If there were a way to check for the extras defined in setup.py, or to link these lists to what is descibed there, that'd be great.

def dtls_missing_modules():
    """Return a list of modules that are missing in order to use the DTLS
    transport, or a false value if everything is present"""

    missing = []

    try:
        from DTLSSocket import dtls # noqa: F401
    except ImportError:
        missing.append("DTLSSocket")

    return missing

def oscore_missing_modules():
    """Return a list of modules that are missing in order to use OSCORE, or a
    false value if everything is present"""
    missing = []
    try:
        import cbor2 # noqa: F401
    except ImportError:
        missing.append('cbor2')
    try:
        import cryptography # noqa: F401
    except ImportError:
        missing.append('cryptography')
    else:
        try:
            from cryptography.hazmat.primitives.ciphers.aead import AESCCM
            AESCCM(b"x" * 16, 8)
        except (cryptography.exceptions.UnsupportedAlgorithm, ImportError):
            missing.append('a version of OpenSSL that supports AES-CCM')
    try:
        import filelock # noqa: F401
    except ImportError:
        missing.append('filelock')

    try:
        import ge25519 # noqa: F401
    except ImportError:
        missing.append('ge25519')

    return missing

def ws_missing_modules():
    """Return a list of modules that are missing in order to user CoAP-over-WS,
    or a false value if everything is present"""

    if is_pyodide:
        return []

    missing = []
    try:
        import websockets # noqa: F401
    except ImportError:
        missing.append('websockets')

    return missing

def linkheader_missing_modules():
    """Return a list of moudles that are missing in order to use link_header
    functionaity (eg. running a resource directory), of a false value if
    everything is present."""
    missing = []
    # The link_header module is now provided in-tree
    return missing

def prettyprint_missing_modules():
    """Return a list of modules that are missing in order to use pretty
    printing (ie. full aiocoap-client)"""
    missing = []
    missing.extend(linkheader_missing_modules())
    try:
        import cbor2 # noqa: F401
    except ImportError:
        missing.append('cbor2')
    try:
        import termcolor # noqa: F401
    except ImportError:
        missing.append('termcolor')
    try:
        import pygments # noqa: F401
    except ImportError:
        missing.append('pygments')
    try:
        import cbor_diag # noqa: F401
    except ImportError:
        missing.append('cbor-diag')
    return missing

missing_module_functions = {
        'dtls': dtls_missing_modules,
        'oscore': oscore_missing_modules,
        'linkheader': linkheader_missing_modules,
        'prettyprint': prettyprint_missing_modules,
        'ws': ws_missing_modules,
        }

__all__ = [
    'get_default_clienttransports',
    'get_default_servertransports',
    'has_reuse_port',
    'dtls_missing_modules',
    'oscore_missing_modules',
    'ws_missing_modules',
    'linkheader_missing_modules',
    'prettyprint_missing_modules',
    'missing_module_functions',
    ]



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import abc
import collections

from .numbers.contentformat import ContentFormat

def _to_minimum_bytes(value):
    return value.to_bytes((value.bit_length() + 7) // 8, 'big')

class OptionType(metaclass=abc.ABCMeta):
    """Interface for decoding and encoding option values

    Instances of :class:`OptionType` are collected in a list in a
    :attr:`.Message.opt` :class:`.Options` object, and provide a translation
    between the CoAP octet-stream (accessed using the
    :meth:`encode()`/:meth:`decode()` method pair) and the interpreted value
    (accessed via the :attr:`value` attribute).

    Note that OptionType objects usually don't need to be handled by library
    users; the recommended way to read and set options is via the Options
    object'sproperties (eg. ``message.opt.uri_path = ('.well-known',
    'core')``)."""

    @abc.abstractmethod
    def __init__(self, number, value):
        """Set the `self.name` and `self.value` attributes"""

    @abc.abstractmethod
    def encode(self):
        """Return the option's value in serialzied form"""

    @abc.abstractmethod
    def decode(self, rawdata):
        """Set the option's value from the bytes in rawdata"""

    def _repr_html_(self):
        import html
        return f"<tt>{html.escape(repr(self))}</tt>"

class StringOption(OptionType):
    """String CoAP option - used to represent string options. Always encoded in
    UTF8 per CoAP specification."""

    def __init__(self, number, value=""):
        self.value = value
        self.number = number

    def encode(self):
        # FIXME: actually, this should be utf8 of the net-unicode form (maybe it is)
        rawdata = self.value.encode('utf-8')
        return rawdata

    def decode(self, rawdata):
        self.value = rawdata.decode('utf-8')

    def __str__(self):
        return self.value

    def _repr_html_(self):
        import html
        return f"<tt>{html.escape(repr(self.value))}</tt>"

class OpaqueOption(OptionType):
    """Opaque CoAP option - used to represent options that just have their
    uninterpreted bytes as value."""

    def __init__(self, number, value=b""):
        self.value = value
        self.number = number

    def encode(self):
        rawdata = self.value
        return rawdata

    def decode(self, rawdata):
        self.value = rawdata

    def __str__(self):
        return repr(self.value)

    def _repr_html_(self):
        return f'<tt>{self.value.hex()}</tt>'

class UintOption(OptionType):
    """Uint CoAP option - used to represent integer options."""

    def __init__(self, number, value=0):
        self.value = value
        self.number = number

    def encode(self):
        return _to_minimum_bytes(int(self.value))

    def decode(self, rawdata):
        self.value = int.from_bytes(rawdata, 'big')

    def __str__(self):
        return str(self.value)

    def _repr_html_(self):
        import html
        return f"<tt>{html.escape(repr(self.value))}</tt>"

class TypedOption(OptionType, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def type(self) -> type:
        """Checked type of the option"""

    def __init__(self, number, value=None):
        self.number = number
        # FIXME when is this ever initialized without value?
        if value is not None:
            self.value = value

    value = property(lambda self: self._value, lambda self, value: self._set_from_opt_value(value))

    def _set_from_opt_value(self, value: object):
        """Convert a value set as ``message.opt.option_name = value`` into the
        stored value. By default, this does an eager isinstance check on the
        value (anticipating that encoding an unsuitable value would otherwise
        fail at a hard-to-debug location)."""
        if not isinstance(value, self.type):
            raise ValueError("Setting values of type %s is not supported on this option" % type(value))
        self._value = value

    def __str__(self):
        return str(self.value)

    def _repr_html_(self):
        if hasattr(self.value, "_repr_html_"):
            return self.value._repr_html_()
        else:
            import html
            return f"<tt>{html.escape(repr(self.value))}</tt>"

class BlockOption(TypedOption):
    """Block CoAP option - special option used only for Block1 and Block2 options.
       Currently it is the only type of CoAP options that has
       internal structure.

       That structure (BlockwiseTuple) covers not only the block options of
       RFC7959, but also the BERT extension of RFC8323. If the reserved size
       exponent 7 is used for purposes incompatible with BERT, the implementor
       might want to look at the context dependent option number
       interpretations which will hopefully be in place for Signaling (7.xx)
       messages by then."""
    class BlockwiseTuple(collections.namedtuple('_BlockwiseTuple', ['block_number', 'more', 'size_exponent'])):
        @property
        def size(self):
            return 2 ** (min(self.size_exponent, 6) + 4)

        @property
        def start(self):
            """The byte offset in the body indicated by block number and size.

            Note that this calculation is only valid for descriptive use and
            Block2 control use. The semantics of block_number and size in
            Block1 control use are unrelated (indicating the acknowledged block
            number in the request Block1 size and the server's preferred block
            size), and must not be calculated using this property in that
            case."""
            return self.block_number * self.size

        @property
        def is_bert(self):
            """True if the exponent is recognized to signal a BERT message."""
            return self.size_exponent == 7

        def is_valid_for_payload_size(self, payloadsize):
            if self.is_bert:
                if self.more:
                    return payloadsize % 1024 == 0
                return True
            else:
                if self.more:
                    return payloadsize == self.size
                else:
                    return payloadsize <= self.size

        def reduced_to(self, maximum_exponent):
            """Return a BlockwiseTuple whose exponent is capped to the given
            maximum_exponent

            >>> initial = BlockOption.BlockwiseTuple(10, 0, 5)
            >>> initial == initial.reduced_to(6)
            True
            >>> initial.reduced_to(3)
            BlockwiseTuple(block_number=40, more=0, size_exponent=3)
            """
            if maximum_exponent >= self.size_exponent:
                return self
            if maximum_exponent == 6 and self.size_exponent == 7:
                return (self.block_number, self.more, 6)
            increasednumber = self.block_number << (min(self.size_exponent, 6) - maximum_exponent)
            return type(self)(increasednumber, self.more, maximum_exponent)

    type = BlockwiseTuple

    def encode(self):
        as_integer = (self.value.block_number << 4) + (self.value.more * 0x08) + self.value.size_exponent
        return _to_minimum_bytes(as_integer)

    def decode(self, rawdata):
        as_integer = int.from_bytes(rawdata, 'big')
        self.value = self.BlockwiseTuple(block_number=(as_integer >> 4), more=bool(as_integer & 0x08), size_exponent=(as_integer & 0x07))

    def _set_from_opt_value(self, value):
        # Casting it through the constructor makes it easy to set the option as
        # `(num, more, sz)` without having to pick the type out of a very long
        # module name
        super()._set_from_opt_value(self.type(*value))

class ContentFormatOption(TypedOption):
    """Type of numeric options whose number has :class:`ContentFormat`
    semantics"""

    type = ContentFormat

    def encode(self):
        return _to_minimum_bytes(int(self.value))

    def decode(self, rawdata):
        as_integer = int.from_bytes(rawdata, 'big')
        self._value = ContentFormat(as_integer)

    def _set_from_opt_value(self, value):
        super()._set_from_opt_value(ContentFormat(value))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from pygments import token, lexers
from pygments.lexer import RegexLexer, bygroups

__all__ = ['LinkFormatLexer']

class LinkFormatLexer(RegexLexer):
    name = "LinkFormatLexer"
    mimetypes = ['application/link-format']

    tokens = {
        'root': [
            ('(<)([^>]*)(>)', bygroups(token.Punctuation, token.Name.Label, token.Punctuation), 'maybe-end')
            ],
        'maybe-end': [
            # Whitespace is not actually allowed, but produced by the pretty printer
            (';\\s*', token.Punctuation, 'attribute'),
            (',\\s*', token.Punctuation, 'root'),
            ],
        'attribute': [
            ('([^,;=]+)((=)("[^"]*"|[^,;"]+))?', bygroups(token.Name.Attribute, None, token.Operator, token.String.Symbol), 'maybe-end'),
            ],
        }

def _register():
    if 'LinkFormatLexer' not in lexers.LEXERS:
        lexers.LEXERS['LinkFormatLexer'] = (
                'aiocoap.util.linkformat_pygments',
                'LinkFormatLexer',
                (),
                (),
                LinkFormatLexer.mimetypes,
                )



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Module containing the CoRE parameters / CoAP Content-Formats registry"""

from __future__ import annotations

from ..util import ExtensibleIntEnum
import warnings

# _raw can be updated from: `curl https://www.iana.org/assignments/core-parameters/content-formats.csv | python3 -c 'import csv, sys; print(list(csv.reader(sys.stdin))[1:])'`

_raw = [
        ['text/plain; charset=utf-8', '', '0', '[RFC2046][RFC3676][RFC5147]'],
        ['Unassigned', '', '1-15', ''],
        ['application/cose; cose-type="cose-encrypt0"', '', '16', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['application/cose; cose-type="cose-mac0"', '', '17', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['application/cose; cose-type="cose-sign1"', '', '18', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['application/ace+cbor', '', '19', '[RFC-ietf-ace-oauth-authz-46]'],
        ['Unassigned', '', '20', ''],
        ['image/gif', '', '21', '[https://www.w3.org/Graphics/GIF/spec-gif89a.txt]'],
        ['image/jpeg', '', '22', '[ISO/IEC 10918-5]'],
        ['image/png', '', '23', '[RFC2083]'],
        ['Unassigned', '', '24-39', ''],
        ['application/link-format', '', '40', '[RFC6690]'],
        ['application/xml', '', '41', '[RFC3023]'],
        ['application/octet-stream', '', '42', '[RFC2045][RFC2046]'],
        ['Unassigned', '', '43-46', ''],
        ['application/exi', '', '47', '["Efficient XML Interchange (EXI) Format 1.0 (Second Edition)", February 2014]'],
        ['Unassigned', '', '48-49', ''],
        ['application/json', '', '50', '[RFC8259]'],
        ['application/json-patch+json', '', '51', '[RFC6902]'],
        ['application/merge-patch+json', '', '52', '[RFC7396]'],
        ['Unassigned', '', '53-59', ''],
        ['application/cbor', '', '60', '[RFC8949]'],
        ['application/cwt', '', '61', '[RFC8392]'],
        ['application/multipart-core', '', '62', '[RFC8710]'],
        ['application/cbor-seq', '', '63', '[RFC8742]'],
        ['Unassigned', '', '64-95', ''],
        ['application/cose; cose-type="cose-encrypt"', '', '96', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['application/cose; cose-type="cose-mac"', '', '97', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['application/cose; cose-type="cose-sign"', '', '98', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['Unassigned', '', '99-100', ''],
        ['application/cose-key', '', '101', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['application/cose-key-set', '', '102', '[RFC-ietf-cose-rfc8152bis-struct-15]'],
        ['Unassigned', '', '103-109', ''],
        ['application/senml+json', '', '110', '[RFC8428]'],
        ['application/sensml+json', '', '111', '[RFC8428]'],
        ['application/senml+cbor', '', '112', '[RFC8428]'],
        ['application/sensml+cbor', '', '113', '[RFC8428]'],
        ['application/senml-exi', '', '114', '[RFC8428]'],
        ['application/sensml-exi', '', '115', '[RFC8428]'],
        ['Unassigned', '', '116-139', ''],
        ['application/yang-data+cbor; id=sid', '', '140', '[RFC9254]'],
        ['Unassigned', '', '141-255', ''],
        ['application/coap-group+json', '', '256', '[RFC7390]'],
        ['application/concise-problem-details+cbor', '', '257', '[RFC-ietf-core-problem-details-08]'],
        ['application/swid+cbor', '', '258', '[RFC-ietf-sacm-coswid-22]'],
        ['Unassigned', '', '259-270', ''],
        ['application/dots+cbor', '', '271', '[RFC9132]'],
        ['application/missing-blocks+cbor-seq', '', '272', '[RFC9177]'],
        ['Unassigned', '', '273-279', ''],
        ['application/pkcs7-mime; smime-type=server-generated-key', '', '280', '[RFC7030][RFC8551][RFC9148]'],
        ['application/pkcs7-mime; smime-type=certs-only', '', '281', '[RFC8551][RFC9148]'],
        ['Unassigned', '', '282-283', ''],
        ['application/pkcs8', '', '284', '[RFC5958][RFC8551][RFC9148]'],
        ['application/csrattrs', '', '285', '[RFC7030][RFC9148]'],
        ['application/pkcs10', '', '286', '[RFC5967][RFC8551][RFC9148]'],
        ['application/pkix-cert', '', '287', '[RFC2585][RFC9148]'],
        ['Unassigned', '', '288-289', ''],
        ['application/aif+cbor', '', '290', '[RFC-ietf-ace-aif-07]'],
        ['application/aif+json', '', '291', '[RFC-ietf-ace-aif-07]'],
        ['Unassigned', '', '292-309', ''],
        ['application/senml+xml', '', '310', '[RFC8428]'],
        ['application/sensml+xml', '', '311', '[RFC8428]'],
        ['Unassigned', '', '312-319', ''],
        ['application/senml-etch+json', '', '320', '[RFC8790]'],
        ['Unassigned', '', '321', ''],
        ['application/senml-etch+cbor', '', '322', '[RFC8790]'],
        ['Unassigned', '', '323-339', ''],
        ['application/yang-data+cbor', '', '340', '[RFC9254]'],
        ['application/yang-data+cbor; id=name', '', '341', '[RFC9254]'],
        ['Unassigned', '', '342-431', ''],
        ['application/td+json', '', '432', '["Web of Things (WoT) Thing Description", May 2019]'],
        ['Unassigned', '', '433-835', ''],
        ['application/voucher-cose+cbor (TEMPORARY - registered 2022-04-12, expires 2023-04-12)', '', '836', '[draft-ietf-anima-constrained-voucher-17]'],
        ['Unassigned', '', '837-1541', ''],
        ['Reserved, do not use', '', '1542-1543', '[OMA-TS-LightweightM2M-V1_0]'],
        ['Unassigned', '', '1544-9999', ''],
        ['application/vnd.ocf+cbor', '', '10000', '[Michael_Koster]'],
        ['application/oscore', '', '10001', '[RFC8613]'],
        ['application/javascript', '', '10002', '[RFC4329]'],
        ['Unassigned', '', '10003-11049', ''],
        ['application/json', 'deflate', '11050', '[RFC8259]'],
        ['Unassigned', '', '11051-11059', ''],
        ['application/cbor', 'deflate', '11060', '[RFC8949]'],
        ['Unassigned', '', '11061-11541', ''],
        ['application/vnd.oma.lwm2m+tlv', '', '11542', '[OMA-TS-LightweightM2M-V1_0]'],
        ['application/vnd.oma.lwm2m+json', '', '11543', '[OMA-TS-LightweightM2M-V1_0]'],
        ['application/vnd.oma.lwm2m+cbor', '', '11544', '[OMA-TS-LightweightM2M-V1_2]'],
        ['Unassigned', '', '11545-19999', ''],
        ['text/css', '', '20000', '[RFC2318]'],
        ['Unassigned', '', '20001-29999', ''],
        ['image/svg+xml', '', '30000', '[https://www.w3.org/TR/SVG/mimereg.html]'],
        ['Unassigned', '', '30001-64999', ''],
        ['Reserved for Experimental Use', '', '65000-65535', '[RFC7252]'],
        ]

def _normalize_media_type(s):
    """Strip out the white space between parameters; doesn't need to fully
    parse the types because it's applied to values of _raw (or to input that'll
    eventually be compared to them and fail)"""
    return s.replace('; ', ';')

class ContentFormat(ExtensibleIntEnum):
    """Entry in the `CoAP Content-Formats registry`__ of the IANA Constrained
    RESTful Environments (Core) Parameters group

    .. __: https://www.iana.org/assignments/core-parameters/core-parameters.xhtml#content-formats

    Known entries have ``.media_type`` and ``.encoding`` attributes:

    >>> ContentFormat(0).media_type
    'text/plain; charset=utf-8'
    >>> int(ContentFormat.by_media_type('text/plain;charset=utf-8'))
    0
    >>> ContentFormat(60)
    <ContentFormat 60, media_type='application/cbor', encoding='identity'>
    >>> ContentFormat(11060).encoding
    'deflate'

    Unknown entries do not have these properties:

    >>> ContentFormat(12345).is_known()
    False
    >>> ContentFormat(12345).media_type                    # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    AttributeError: ...

    Only a few formats are available as attributes for easy access. Their
    selection and naming are arbitrary and biased. The remaining known types
    are available through the :meth:`by_media_type` class method.
    >>> ContentFormat.TEXT
    <ContentFormat 0, media_type='text/plain; charset=utf-8', encoding='identity'>

    A convenient property of ContentFormat is that any known content format is
    true in a boolean context, and thus when used in alternation with None, can
    be assigned defaults easily:

    >>> requested_by_client = ContentFormat.TEXT
    >>> int(requested_by_client) # Usually, this would always pick the default
    0
    >>> used = requested_by_client or ContentFormat.LINKFORMAT
    >>> assert used == ContentFormat.TEXT
    """

    @classmethod
    def by_media_type(cls, media_type: str, encoding: str = 'identity') -> ContentFormat:
        """Produce known entry for a known media type (and encoding, though
        'identity' is default due to its prevalence), or raise KeyError."""
        return cls._by_mt_encoding[(_normalize_media_type(media_type), encoding)]

    def is_known(self):
        return hasattr(self, "media_type")

    @classmethod
    def _rehash(cls):
        """Update the class's cache of known media types

        Run this after having created entries with media type and encoding that
        should be found later on."""
        cls._by_mt_encoding = {(_normalize_media_type(c.media_type), c.encoding): c for c in cls._value2member_map_.values()}

    def __repr__(self):
        return "<%s %d%s>" % (type(self).__name__, self, ', media_type=%r, encoding=%r' % (self.media_type, self.encoding) if self.is_known() else "")

    def __bool__(self):
        return True

    def _repr_html_(self):
        # The name with title thing isn't too pretty for these ones
        if self.is_known():
            import html
            return f"""<abbr title="Content format {int(self)}{', named ContentFormat.' + html.escape(self.name) if hasattr(self, 'name') else ''}">{html.escape(self.media_type)}{'@' + self.encoding if self.encoding != 'identity' else ''}</abbr>"""
        else:
            return f"""<abbr title="Unknown content format">{int(self)}</abbr>"""

    TEXT = 0
    LINKFORMAT = 40
    OCTETSTREAM = 42
    JSON = 50
    CBOR = 60
    SENML = 112

for (_mt, _enc, _i, _source) in _raw:
    if _mt in ["Reserved for Experimental Use", "Reserved, do not use", "Unassigned"]:
        continue
    _mt, _, _ = _mt.partition(' (TEMPORARY')
    _cf = ContentFormat(int(_i))
    _cf.media_type = _mt
    _cf.encoding = _enc or "identity"
ContentFormat._rehash()


class _MediaTypes:
    """Wrapper to provide a media_types indexable object as was present up to
    0.4.2"""
    def __getitem__(self, content_format):
        warnings.warn("media_types is deprecated, please use aiocoap.numbers.ContentFormat", DeprecationWarning, stacklevel=2)
        if content_format is None:
            # That was a convenient idiom to short-circuit through, but would
            # fail through the constructor
            raise KeyError(None)

        cf = ContentFormat(content_format)
        if cf.is_known():
            return _normalize_media_type(cf.media_type)
        else:
            raise KeyError(content_format)

    def get(self, content_format, default=None):
        warnings.warn("media_types is deprecated, please use aiocoap.numbers.ContentFormat", DeprecationWarning, stacklevel=2)
        try:
            return self[content_format]
        except KeyError:
            return default

class _MediaTypesRev:
    """Wrapper to provide a media_types_rev indexable object as was present up
    to 0.4.2"""
    def __getitem__(self, name):
        warnings.warn("media_types_rev is deprecated, please use aiocoap.numbers.ContentFormat", DeprecationWarning, stacklevel=2)
        if name == 'text/plain':
            # deprecated alias. Kept alive for scripts like
            # https://gitlab.f-interop.eu/f-interop-contributors/ioppytest/blob/develop/automation/coap_client_aiocoap/automated_iut.py
            # that run aiocoap-client with text/plain as an argument.
            name = 'text/plain;charset=utf-8'
        return int(ContentFormat.by_media_type(name))

    def get(self, name, default=None):
        warnings.warn("media_types_rev is deprecated, please use aiocoap.numbers.ContentFormat", DeprecationWarning, stacklevel=2)
        try:
            return self[name]
        except KeyError:
            return default



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Common options of aiocoap command line utilities

Unlike those in :mod:`aiocoap.util.cli`, these are particular to aiocoap
functionality.

Typical use is like this::

>>> p = argparse.ArgumentParser()
>>> p.add_argument('--foo')                         # doctest: +ELLIPSIS
_...
>>> add_server_arguments(p)
>>> opts = p.parse_args(['--bind', '[::1]:56830', '--foo=bar'])

You can then either pass opts directly to
:func:`server_context_from_arguments`, or split up the arguments::

>>> server_opts = extract_server_arguments(opts)
>>> opts
Namespace(foo='bar')

Then, server_opts can be passed to `server_context_from_arguments`.
"""

import sys
import argparse
import json
from pathlib import Path

from ..util import hostportsplit
from ..protocol import Context
from ..credentials import CredentialsMap

class _HelpBind(argparse.Action):
    def __init__(self, *args, **kwargs):
        kwargs['nargs'] = 0
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        print("The --bind option can take either of the following formats:"
                "\n    :port -- bind to a given port on all available interfaces"
                "\n    host -- bind to default ports on a given host name (can also be an IP address; IPv6 addresses need to be in square brackets)"
                "\n    host:port -- bind only to a specific port on a given host"
                "\n\nBy default, the server will bind to all available addressess and protocols on the respective default ports."
                "\nIf a port is specified, and (D)TLS support is available, those protocols will be bound to one port higher (as are the default ports, 5683 for CoAP and 5684 for CoAP over (D)TLS)."
                "\n", file=sys.stderr)
        parser.exit()

def add_server_arguments(parser):
    """Add the --bind option to an argparse parser"""

    def hostportsplit_helper(arg):
        """Wrapper around hostportsplit that gives better error messages than
        'invalid hostportsplit value'"""

        if arg.isnumeric():
            raise parser.error("Invalid argument to --bind."
                    f" Did you mean --bind :{arg}?")

        try:
            return hostportsplit(arg)
        except ValueError:
            raise parser.error("Invalid argument to --bind." +
                    " Did you mean --bind '[%s]'?" % arg
                    if arg.count(':') >= 2 and '[' not in arg
                    else " See --help-bind for details.")

    parser.add_argument('--bind', help="Host and/or port to bind to (see --help-bind for details)", type=hostportsplit_helper, default=None)

    parser.add_argument('--credentials', help="JSON file pointing to credentials for the server's identity/ies.", type=Path)

    # These are to be eventually migrated into credentials
    parser.add_argument('--tls-server-certificate', help="TLS certificate (chain) to present to connecting clients (in PEM format)", metavar="CRT")
    parser.add_argument('--tls-server-key', help="TLS key to load that supports the server certificate", metavar="KEY")

    parser.add_argument('--help-bind', help=argparse.SUPPRESS, action=_HelpBind)

def extract_server_arguments(namespace):
    """Given the output of .parse() on a ArgumentParser that had
    add_server_arguments called with it, remove the resulting option in-place
    from namespace and return them in a separate namespace."""

    server_arguments = type(namespace)()
    server_arguments.bind = namespace.bind
    server_arguments.tls_server_certificate = namespace.tls_server_certificate
    server_arguments.tls_server_key = namespace.tls_server_key
    server_arguments.credentials = namespace.credentials

    del namespace.bind
    del namespace.tls_server_certificate
    del namespace.tls_server_key
    del namespace.credentials
    del namespace.help_bind

    return server_arguments

async def server_context_from_arguments(site, namespace, **kwargs):
    """Create a bound context like
    :meth:`.aiocoap.Context.create_server_context`, but take the bind and TLS
    settings from a namespace returned from an argparse parser that has had
    :func:`add_server_arguments` run on it.
    """

    if namespace.tls_server_certificate:
        import ssl

        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=namespace.tls_server_certificate, keyfile=namespace.tls_server_key)
        ssl_context.set_alpn_protocols(["coap"])
        if hasattr(ssl_context, 'sni_callback'): # starting python 3.7
            ssl_context.sni_callback = lambda obj, name, context: setattr(obj, "indicated_server_name", name)
    else:
        ssl_context = None

    if namespace.credentials:
        server_credentials = CredentialsMap()
        server_credentials.load_from_dict(json.load(namespace.credentials.open('rb')))

        # FIXME: could be non-OSCORE as well -- can we build oscore_sitewrapper
        # in such a way it only depends on the OSCORE dependencies if there are
        # actual identities present?
        from aiocoap.oscore_sitewrapper import OscoreSiteWrapper
        site = OscoreSiteWrapper(site, server_credentials)
    else:
        server_credentials = None

    return await Context.create_server_context(site, namespace.bind, _ssl_context=ssl_context, server_credentials=server_credentials, **kwargs)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""A simple file server that serves the contents of a given directory in a
read-only fashion via CoAP. It provides directory listings, and guesses the
media type of files it serves.

It follows the conventions set out for the [kitchen-sink fileserver],
optionally with write support, with some caveats:

* There are some time-of-check / time-of-use race conditions around the
  handling of ETags, which could probably only be resolved if heavy file system
  locking were used. Some of these races are a consequence of this server
  implementing atomic writes through renames.

  As long as no other processes access the working area, and aiocoap is run
  single threaded, the races should not be visible to CoAP users.

* ETags are constructed based on information in the file's (or directory's)
  `stat` output -- this avoids reaing the whole file on overwrites etc.

  This means that forcing the MTime to stay constant across a change would
  confuse clients.

* While GET requests on files are served block by block (reading only what is
  being requested), PUT operations are spooled in memory rather than on the
  file system.

* Directory creation and deletion is not supported at the moment.

[kitchen-sink fileserver]: https://www.ietf.org/archive/id/draft-amsuess-core-coap-kitchensink-00.html#name-coap
"""

import argparse
import asyncio
from pathlib import Path
import logging
from stat import S_ISREG, S_ISDIR
import mimetypes
import tempfile
import hashlib

import aiocoap
import aiocoap.error as error
import aiocoap.numbers.codes as codes
from aiocoap.resource import Resource
from aiocoap.util.cli import AsyncCLIDaemon
from aiocoap.cli.common import (add_server_arguments,
    server_context_from_arguments, extract_server_arguments)
from aiocoap.resourcedirectory.client.register import Registerer
from ..util.asyncio import py38args

class InvalidPathError(error.ConstructionRenderableError):
    code = codes.BAD_REQUEST

class TrailingSlashMissingError(error.ConstructionRenderableError):
    code = codes.BAD_REQUEST
    message = "Error: Not a file (add trailing slash)"

class AbundantTrailingSlashError(error.ConstructionRenderableError):
    code = codes.BAD_REQUEST
    message = "Error: Not a directory (strip the trailing slash)"

class NoSuchFile(error.NotFound): # just for the better error msg
    message = "Error: File not found!"

class PreconditionFailed(error.ConstructionRenderableError):
    code = codes.PRECONDITION_FAILED

class FileServer(Resource, aiocoap.interfaces.ObservableResource):
    # Resource is only used to give the nice render_xxx methods

    def __init__(self, root, log, *, write=False):
        super().__init__()
        self.root = root
        self.log = log
        self.write = write

        self._observations = {} # path -> [last_stat, [callbacks]]

    # While we don't have a .well-known/core resource that would need this, we
    # still allow registration at an RD and thus need something in here.
    #
    # As we can't possibly register all files in here, we're just registering a
    # single link to the index.
    def get_resources_as_linkheader(self):
        # Resource type indicates draft-amsuess-core-coap-kitchensink-00 file
        # service, might use registered name later
        return '</>;ct=40;rt="tag:chrysn@fsfe.org,2022:fileserver"'

    async def check_files_for_refreshes(self):
        while True:
            await asyncio.sleep(10)

            for path, data in list(self._observations.items()):
                last_stat, callbacks = data
                if last_stat is None:
                    continue # this hit before the original response even triggered
                try:
                    new_stat = path.stat()
                except Exception:
                    new_stat = False
                def relevant(s):
                    return (s.st_ino, s.st_dev, s.st_size, s.st_mtime, s.st_ctime)
                if relevant(new_stat) != relevant(last_stat):
                    self.log.info("New stat for %s", path)
                    data[0] = new_stat
                    for cb in callbacks:
                        cb()

    def request_to_localpath(self, request):
        path = request.opt.uri_path
        if any('/' in p or p in ('.', '..') for p in path):
            raise InvalidPathError()

        return self.root / "/".join(path)

    async def needs_blockwise_assembly(self, request):
        if request.code != codes.GET:
            return True
        if not request.opt.uri_path or request.opt.uri_path[-1] == '' or \
                request.opt.uri_path == ('.well-known', 'core'):
            return True
        # Only GETs to non-directory access handle it explicitly
        return False

    @staticmethod
    def hash_stat(stat):
        # The subset that the author expects to (possibly) change if the file changes
        data = (stat.st_mtime_ns, stat.st_ctime_ns, stat.st_size)
        return hashlib.sha256(repr(data).encode('ascii')).digest()[:8]

    async def render_get(self, request):
        if request.opt.uri_path == ('.well-known', 'core'):
            return aiocoap.Message(
                    payload=str(self.get_resources_as_linkheader()).encode('utf8'),
                    content_format=40
                    )

        path = self.request_to_localpath(request)
        try:
            st = path.stat()
        except FileNotFoundError:
            raise NoSuchFile()

        etag = self.hash_stat(st)

        if etag in request.opt.etags:
            response = aiocoap.Message(code=codes.VALID)
        else:
            if S_ISDIR(st.st_mode):
                response = await self.render_get_dir(request, path)
            elif S_ISREG(st.st_mode):
                response = await self.render_get_file(request, path)

        response.opt.etag = etag
        return response

    async def render_put(self, request):
        if not self.write:
            return aiocoap.Message(code=codes.FORBIDDEN)

        if not request.opt.uri_path or not request.opt.uri_path[-1]:
            # Attempting to write to a directory
            return aiocoap.Message(code=codes.BAD_REQUEST)

        path = self.request_to_localpath(request)

        if request.opt.if_none_match:
            # FIXME: This is locally a race condition; files could be created
            # in the "x" mode, but then how would writes to them be made
            # atomic?
            if path.exists():
                raise PreconditionFailed()

        if request.opt.if_match and b"" not in request.opt.if_match:
            # FIXME: This is locally a race condition; not sure how to prevent
            # that.
            try:
                st = path.stat()
            except FileNotFoundError:
                # Absent file in particular doesn't have the expected ETag
                raise PreconditionFailed()
            if self.hash_stat(st) not in request.opt.if_match:
                raise PreconditionFailed()

        # Is there a way to get both "Use umask for file creation (or the
        # existing file's permissions)" logic *and* atomic file creation on
        # portable UNIX? If not, all we could do would be emulate the logic of
        # just opening the file (by interpreting umask and the existing file's
        # permissions), and that fails horrobly if there are ACLs in place that
        # bites rsync in https://bugzilla.samba.org/show_bug.cgi?id=9377.
        #
        # If there is not, secure temporary file creation is as good as
        # anything else.
        with tempfile.NamedTemporaryFile(dir=path.parent, delete=False) as spool:
            spool.write(request.payload)
            temppath = Path(spool.name)
        try:
            temppath.rename(path)
        except Exception:
            temppath.unlink()
            raise

        st = path.stat()
        etag = self.hash_stat(st)

        return aiocoap.Message(code=codes.CHANGED, etag=etag)

    async def render_delete(self, request):
        if not self.write:
            return aiocoap.Message(code=codes.FORBIDDEN)

        if not request.opt.uri_path or not request.opt.uri_path[-1]:
            # Deleting directories is not supported as they can't be created
            return aiocoap.Message(code=codes.BAD_REQUEST)

        path = self.request_to_localpath(request)

        if request.opt.if_match and b"" not in request.opt.if_match:
            # FIXME: This is locally a race condition; not sure how to prevent
            # that.
            try:
                st = path.stat()
            except FileNotFoundError:
                # Absent file in particular doesn't have the expected ETag
                raise NoSuchFile()
            if self.hash_stat(st) not in request.opt.if_match:
                raise PreconditionFailed()

        try:
            path.unlink()
        except FileNotFoundError:
            raise NoSuchFile()

        return aiocoap.Message(code=codes.DELETED)

    async def render_get_dir(self, request, path):
        if request.opt.uri_path and request.opt.uri_path[-1] != '':
            raise TrailingSlashMissingError()

        self.log.info("Serving directory %s", path)

        response = ""
        for f in path.iterdir():
            rel = f.relative_to(self.root)
            if f.is_dir():
                response += "</%s/>;ct=40," % rel
            else:
                response += "</%s>," % rel
        return aiocoap.Message(payload=response[:-1].encode('utf8'), content_format=40)

    async def render_get_file(self, request, path):
        if request.opt.uri_path and request.opt.uri_path[-1] == '':
            raise AbundantTrailingSlashError()

        self.log.info("Serving file %s", path)

        block_in = request.opt.block2 or aiocoap.optiontypes.BlockOption.BlockwiseTuple(0, 0, 6)

        with path.open('rb') as f:
            f.seek(block_in.start)
            data = f.read(block_in.size + 1)

        if path in self._observations and self._observations[path][0] is None:
            # FIXME this is not *completely* precise, as it might mean that in
            # a (Observation 1 established, check loop run, file modified,
            # observation 2 established) situation, observation 2 could receive
            # a needless update on the next check, but it's simple and errs on
            # the side of caution.
            self._observations[path][0] = path.stat()

        guessed_type, _ = mimetypes.guess_type(str(path))

        block_out = aiocoap.optiontypes.BlockOption.BlockwiseTuple(block_in.block_number, len(data) > block_in.size, block_in.size_exponent)
        content_format = None
        if guessed_type is not None:
            try:
                content_format = aiocoap.numbers.ContentFormat.by_media_type(guessed_type)
            except KeyError:
                if guessed_type and guessed_type.startswith('text/'):
                    content_format = aiocoap.numbers.ContentFormat.TEXT
        return aiocoap.Message(
                payload=data[:block_in.size],
                block2=block_out,
                content_format=content_format,
                observe=request.opt.observe
                )

    async def add_observation(self, request, serverobservation):
        path = self.request_to_localpath(request)

        # the actual observable flag will only be set on files anyway, the
        # library will cancel the file observation accordingly if the requested
        # thing is not actually a file -- so it can be done unconditionally here

        last_stat, callbacks = self._observations.setdefault(path, [None, []])
        cb = serverobservation.trigger
        callbacks.append(cb)
        serverobservation.accept(lambda self=self, path=path, cb=cb: self._observations[path][1].remove(cb))

class FileServerProgram(AsyncCLIDaemon):
    async def start(self):
        logging.basicConfig()

        self.registerer = None

        p = self.build_parser()

        opts = p.parse_args()
        server_opts = extract_server_arguments(opts)

        await self.start_with_options(**vars(opts), server_opts=server_opts)

    @staticmethod
    def build_parser():
        p = argparse.ArgumentParser(description=__doc__)
        p.add_argument("-v", "--verbose", help="Be more verbose (repeat to debug)", action='count', dest="verbosity", default=0)
        p.add_argument("--register", help="Register with a Resource directory", metavar='RD-URI', nargs='?', default=False)
        p.add_argument("--write", help="Allow writes by any user", action='store_true')
        p.add_argument("path", help="Root directory of the server", nargs="?", default=".", type=Path)

        add_server_arguments(p)

        return p

    async def start_with_options(self, path, verbosity=0, register=False,
            server_opts=None, write=False):
        log = logging.getLogger('fileserver')
        coaplog = logging.getLogger('coap-server')

        if verbosity == 1:
            log.setLevel(logging.INFO)
        elif verbosity == 2:
            log.setLevel(logging.DEBUG)
            coaplog.setLevel(logging.INFO)
        elif verbosity >= 3:
            log.setLevel(logging.DEBUG)
            coaplog.setLevel(logging.DEBUG)

        server = FileServer(path, log, write=write)
        if server_opts is None:
            self.context = await aiocoap.Context.create_server_context(server)
        else:
            self.context = await server_context_from_arguments(server, server_opts)

        self.refreshes = asyncio.create_task(
                server.check_files_for_refreshes(),
                **py38args(name="Refresh on %r" % (path,))
                )

        if register is not False:
            if register is not None and register.count('/') != 2:
                log.warn("Resource directory does not look like a host-only CoAP URI")

            self.registerer = Registerer(self.context, rd=register, lt=60)

            if verbosity == 2:
                self.registerer.log.setLevel(logging.INFO)
            elif verbosity >= 3:
                self.registerer.log.setLevel(logging.DEBUG)

    async def shutdown(self):
        if self.registerer is not None:
            await self.registerer.shutdown()
        self.refreshes.cancel()
        await self.context.shutdown()

# used by doc/aiocoap_index.py
build_parser = FileServerProgram.build_parser

if __name__ == "__main__":
    FileServerProgram.sync_main()
